
#ifndef __HSCAL_BLACS_BIBECOMB_HSCALFUNC_HH
#define __HSCAL_BLACS_BIBECOMB_HSCALFUNC_HH

#include "../../../../include/scalapack.h"

/*
 *  The bidirectional exchange topology (BE) is specialized for dealing with
 *  case where all nodes participating in the operation need to
 *  receive the answer.  It works best when # of nodes is some even
 *  power of two.  This topology is based on an algorithm presented by
 *  Robert van de Geijn, et al.
 */
void hscal_BI_BeComb_nocomm(
     HSCALBLACSCONTEXT *ctxt, int NC, int N, HSCALVVFUNPTR Xvvop,
     int nprow, int npcol, int myrow, int mycol, 
     double *tcomp, int *tcomm
)
/*
 *  -- V1.1ALPHA (test version) BLACS routine --
 *  University of Tennessee, October 1, 1995
 *  Written by Clint Whaley.
 *
 *  Purpose
 *  =======
 *  Perform a element-by-element combine on vectors.
 *  The answer will be left on all participating processes.  Since this method
 *  uses a hypercube communication pattern, the number of nodes participating
 *  in the operation must be a power of 2 for it to perform efficiently.
 *
 *  Arguments
 *  =========
 *  CTXT    (input) pointer to BLACSCONTEXT
 *          The BLACS context where operation is taking place.
 *
 *  BP      (input/output) pointer to BLACBUFF.
 *          BLACBUFF is a special data type used by the BLACS to control
 *          buffers and the asynchronous operations coming out of them.
 *          This BLACBUFF should have a buffer who's first N elements
 *          contain the data to be combined. Additional space may be
 *          required, depending upon what combine is being performed.
 *
 *  BP2     (workspace) pointer to BLACBUFF.
 *          This BLACBUFF is used to receive information for combining with
 *          this process's information.
 *
 *  N       (input) int
 *          The number of elements in the vector to be combined.
 *
 *  Xvvop   (input) pointer to typed operation function
 *          Points to a typed function which performs the required operation
 *          (e.g. summation) on the two N-element vectors.
 *
 * ------------------------------------------------------------------------
 */
{
   int Np, Iam, dest, np2, bit, ierr;
   char scope = ctxt->scp->scope;

   if (N == 0)
   {
      return;
   }

   HSCAL_DEBUG(__FILE__, __FUNCTION__,__LINE__, "N %d", N);

   Np = ctxt->scp->Np;
   if (Np < 2) return;
   Iam = ctxt->scp->Iam;

   for (np2=4; np2 < Np; np2 <<= 1);
   if (np2 > Np) np2 >>= 1;

   if (np2 != Np)
   {
      dest = (Iam ^ np2);
      if (Iam >= np2)  /* I'm node beyond power of 2 */
      {
         if (tcomm != NULL)
         {
            int idest, jdest;

            switch (scope)
            {
            case 'c':
            {
                TCOMM(myrow, mycol, dest, mycol, nprow, npcol) += N;
                break;
            }
            case 'r':
            {
                TCOMM(myrow, mycol, myrow, dest, nprow, npcol) += N;
                break;
            }
            case 'a':
            {
                idest = dest/npcol;
                jdest = dest%npcol;
                TCOMM(myrow, mycol, idest, jdest, nprow, npcol) += N;
                break;
            }
            default:
                break;
            }
         }
      }
      else if (Iam < (Np^np2))  /* need to fan in contents of */
      {                         /* non-power of 2 nodes */
         Xvvop(
           NC,
           nprow, npcol, myrow, mycol, tcomp
         );
      }
   }

   if (Iam < np2)
   {
      for (bit=1; (bit ^ np2); bit <<= 1)
      {
         dest = Iam ^ bit;
         if (tcomm != NULL)
         {
            int idest, jdest;
            switch (scope)
            {
            case 'c':
            {
                TCOMM(myrow, mycol, dest, mycol, nprow, npcol) += N;
                break;
            }
            case 'r':
            {
                TCOMM(myrow, mycol, myrow, dest, nprow, npcol) += N;
                break;
            }
            case 'a':
            {
                idest = dest/npcol;
                jdest = dest%npcol;
                TCOMM(myrow, mycol, idest, jdest, nprow, npcol) += N;
                break;
            }
            default:
                break;
            }
         }
         Xvvop(
           NC,
           nprow, npcol, myrow, mycol, tcomp
         );
      }
/*
 *  For nodes that are not part of the hypercube proper, we must
 *  send data back.
 */
      if (Iam < (Np^np2)) {
         int dest = (Iam ^ np2);
         if (tcomm != NULL)
         {
            int idest, jdest;
            switch (scope)
            {
            case 'c':
            {
                TCOMM(myrow, mycol, dest, mycol, nprow, npcol) += N;
                break;
            }
            case 'r':
            {
                TCOMM(myrow, mycol, myrow, dest, nprow, npcol) += N;
                break;
            }
            case 'a':
            {
                idest = dest/npcol;
                jdest = dest%npcol;
                TCOMM(myrow, mycol, idest, jdest, nprow, npcol) += N;
                break;
            }
            default:
                break;
            }
         }
      }
   }  /* end if (nodes inside power of 2) */
}

#endif
