
#ifndef __HSCAL_BLACS_BITREEBS_HSCALFUNC_HH
#define __HSCAL_BLACS_BITREEBS_HSCALFUNC_HH

#include "../../../../include/scalapack.h"

/*
 *  Tree_bs/br is a algorithm that does a broadcast send/recv such that the
 *  communication pattern is a tree with an arbitrary number of branches.
 *  The following two pairs of graphs give different ways of viewing the same
 *  algorithm.  The first pair shows the trees as they should be visualized
 *  when examining the algorithm.  The second pair are isomorphic graphs of
 *  of the first, which show the actual pattern of data movement.

 *  Note that a tree broadcast with NBRANCHES = 2 is isomorphic with a
 *  hypercube broadcast (however, it does not require the nodes be a
 *  power of two to work).
 *
 *    TREE BROADCAST, NBRANCHES = 2     *    TREE BROADCAST, NBRANCHES = 3
 *                                      *
 * i=4   &______________                *
 *       |                            *
 * i=2   &______         &______        * i=3     &______________________
 *       |             |            *         |                   
 * i=1   &__     &__     &__     &__    * i=1     &______     &______     &__
 *       |     |     |     |    *         |       |       | 
 *       0   1   2   3   4   5   6   7  *         0   1   2   3   4   5   6   7
 *
 *
 *          ISOMORPHIC GRAPHS OF ABOVE, SHOWN IN MORE FAMILIAR TERMS:
 *
 *                0                                           0
 *       _________|_________                       ___________|____________
 *      /         |                             /           |      |    
 *     4          2          1                   6            3      2      1
 *    /         |                              |           /
 *   6   5        3                              7          4   5
 *   |
 *   7
 */
void hscal_BI_TreeBS_nocomm(
     HSCALBLACSCONTEXT *ctxt, int NC, int N, HSCALSDPTR send, int nbranches,
     int nprow, int npcol, int myrow, int mycol, 
     double *tcomp, int *tcomm
)
{
   int Np, Iam, i, j;
   int destdist;	/* the distance of the destination node */
   char scope = ctxt->scp->scope;

   if (N == 0)
   {
      return;
   }

   HSCAL_DEBUG(__FILE__, __FUNCTION__,__LINE__, "N %d nbranches %d", N, nbranches);

   Np = ctxt->scp->Np;
   if (Np < 2) return;
   Iam = ctxt->scp->Iam;
   for (i=nbranches; i < Np; i*=nbranches);
   for (i /= nbranches; i > 0; i /= nbranches)
   {
      j = 1;
      do
      {
	 destdist = i*j;
	 if (destdist < Np)
         {
            int dest = (destdist+Iam)%Np;
            if (tcomm != NULL)
            {
               int idest, jdest;

               switch (scope)
               {
               case 'c':
               {      
                   send(
                     N, dest, mycol,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                   );
                   break;
               }
               case 'r':
               {
                   send(
                     N, myrow, dest,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                   );
                   break;
               }
               case 'a':
               {
                   idest = dest/npcol;
                   jdest = dest%npcol;
                   send(
                     N, idest, jdest,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                   );
                   break;
               }
               default:
                   break;
               }
            }
         }
      }
      while(++j < nbranches);
   }
} /* end BI_TreeBS */

#endif
