
#ifndef __HSCAL_BLACS_CCGAMX2D_HSCALFUNC_HH
#define __HSCAL_BLACS_CCGAMX2D_HSCALFUNC_HH

#include <stdlib.h>

#include "../../../include/scalapack.h"

void hscal_Ccgamx2d(
     HSCALBLACSCONTEXT *ctxt, char *scope, char *top, int m, int n,
     int lda, int *rA, int *cA, int ldia, int rdest, int cdest,
     int nprow, int npcol, int myrow, int mycol, 
     double *tcomp, int *tcomm
)
/*
 *  -- V1.1 BLACS routine --
 *  University of Tennessee, May 1, 1996
 *  Written by Clint Whaley.
 *
 *  Purpose
 *  =======
 *  Combine amx operation for complex rectangular matrices.
 *
 *  Arguments
 *  =========
 *
 *  ConTxt  (input) Ptr to int
 *          Index into MyConTxts00 (my contexts array).
 *
 *  SCOPE   (input) Ptr to char
 *          Limit the scope of the operation.
 *          = 'R' :   Operation is performed by a process row.
 *          = 'C' :   Operation is performed by a process column.
 *          = 'A' :   Operation is performed by all processes in grid.
 *
 *  TOP     (input) Ptr to char
 *          Controls fashion in which messages flow within the operation.
 *
 *  M       (input) Ptr to int
 *          The number of rows of the matrix A.  M >= 0.
 *
 *  N       (input) Ptr to int
 *          The number of columns of the matrix A.  N >= 0.
 *
 *  A       (output) Ptr to complex two dimensional array
 *          The m by n matrix A.  Fortran77 (column-major) storage
 *          assumed.
 *
 *  LDA     (input) Ptr to int
 *          The leading dimension of the array A.  LDA >= M.
 *
 *  RA      (output) Integer Array, dimension (LDIA, N)
 *          Contains process row that the amx of each element
 *          of A was found on: i.e., rA(1,2) contains the process
 *          row that the amx of A(1,2) was found on.
 *          Values are left on process {rdest, cdest} only, others
 *          may be modified, but not left with interesting data.
 *          If rdest == -1, then result is left on all processes in scope.
 *          If LDIA == -1, this array is not accessed, and need not exist.
 *
 *  CA      (output) Integer Array, dimension (LDIA, N)
 *          Contains process column that the amx of each element
 *          of A was found on: i.e., cA(1,2) contains the process
 *          column that the max/min of A(1,2) was found on.
 *          Values are left on process {rdest, cdest} only, others
 *          may be modified, but not left with interesting data.
 *          If rdest == -1, then result is left on all processes in scope.
 *          If LDIA == -1, this array is not accessed, and need not exist.
 *
 *  LDIA    (input) Ptr to int
 *          If (LDIA == -1), then the arrays RA and CA are not accessed.
 *          ELSE leading dimension of the arrays RA and CA.  LDIA >= M.
 *
 *  RDEST   (input) Ptr to int
 *          The process row of the destination of the amx.
 *          If rdest == -1, then result is left on all processes in scope.
 *
 *  CDEST   (input) Ptr to int
 *          The process column of the destination of the amx.
 *          If rdest == -1, then CDEST ignored.
 *
 * ------------------------------------------------------------------------
 */
{
/*
 *  Variable Declarations
 */
   HSCALVVFUNPTR vvop;
   char ttop, tscope;
   int i, j, N, dest, idist, length, tlda, tldia, trdest, ierr;
   int contiguous = 0;
   BI_DistType *dist, mydist;

 /*
  * Fill up the uninitialized fields
  */
   ctxt->rscp.Iam = mycol;
   ctxt->cscp.Iam = myrow;
   ctxt->pscp.Iam = ctxt->ascp.Iam = myrow*npcol+mycol;

   ttop = F2C_CharTrans(top);
   ttop = Mlowcase(ttop);
   tscope = F2C_CharTrans(scope);
   tscope = Mlowcase(tscope);
/*
 *  If the user has set the default combine topology, use it instead of
 *  BLACS default
 */
   if (ttop == ' ') ttop = DefCombTop;
   if (Mpval(cdest) == -1) trdest = -1;
   else trdest = Mpval(rdest);

   if (Mpval(lda) >= Mpval(m)) tlda = Mpval(lda);
   else tlda = Mpval(m);
   if (Mpval(ldia) < Mpval(m)) tldia = Mpval(m);
   else tldia = Mpval(ldia);

   switch(tscope)
   {
   case 'r':
      ctxt->scp = &ctxt->rscp;
      if (trdest == -1) dest = -1;
      else dest = Mpval(cdest);
      break;
   case 'c':
      ctxt->scp = &ctxt->cscp;
      dest = trdest;
      break;
   case 'a':
      ctxt->scp = &ctxt->ascp;
      if (trdest == -1) dest = -1;
      else dest = Mvkpnum(ctxt, trdest, Mpval(cdest));
      break;
   default:
      BI_BlacsErr_nocomm(ctxt, __LINE__, __FILE__, "Unknown scope '%c'",
                  tscope);
   }


/*
 * It's not defined how MPI reacts to 0 element reductions, so use BLACS 1-tree
 * topology if we've got one
 */
   if (ttop == ' ')
      if ( (Mpval(m) < 1) || (Mpval(n) < 1) || (ctxt->TopsRepeat) ) ttop = '1';
   N = Mpval(m) * Mpval(n);

/*
 * If process who has amx is to be communicated, must set up distance
 * vector after value vector
 */
   if (Mpval(ldia) != -1)
   {
      vvop = hscal_BI_cvvamx_nocomm;
      length = N * sizeof(SCOMPLEX);
      i = length % sizeof(BI_DistType);  /* ensure dist vec aligned correctly */
      if (i) length += sizeof(BI_DistType) - i;
      idist = length;
      length += N * sizeof(BI_DistType);
/*
 *    For performance, insist second buffer is at least 8-byte aligned
 */
      j = 8;
      if (sizeof(SCOMPLEX) > j) j = sizeof(SCOMPLEX);
      i = length % j;
      if (i) length += j - i;
      i = 2 * length;

      hscal_BI_cmvcopy_nocomm(
           Mpval(m), Mpval(n), tlda,
           nprow, npcol, myrow, mycol, tcomp
      );
/*
 *    Fill in distance vector
 */
      if (dest == -1) mydist = ctxt->scp->Iam;
      else mydist = (ctxt->scp->Np + ctxt->scp->Iam - dest) % ctxt->scp->Np;

      dist = (BI_DistType *)malloc(sizeof(BI_DistType)*N);
      if (dist == NULL) 
      {
         BI_BlacsErr_nocomm(
           ctxt, __LINE__, __FILE__, "No memory to allocate distance vector 'dist'"
         );
         return;
      }
      for (i=0; i < N; i++) 
      {
         dist[i] = mydist;
      }
   }
   else
   {
      vvop = hscal_BI_cvvamx2_nocomm;
      length = N * sizeof(SCOMPLEX);
/*
 *    If A is contiguous, we can use it as one of our buffers
 */
      if ( (Mpval(m) == tlda) || (Mpval(n) == 1) )
      {
         contiguous = 1;
      }
      else
      {
         hscal_BI_cmvcopy_nocomm(
              Mpval(m), Mpval(n), tlda,
              nprow, npcol, myrow, mycol, tcomp
         );
      }
   }

   switch(ttop)
   {
   case ' ':         /* use MPI's reduction by default */
      i = 1;
      if (Mpval(ldia) == -1)
      {
         vvop = hscal_BI_cMPI_amx2_nocomm;
      }
      else
      {
         vvop = hscal_BI_cMPI_amx_nocomm;
      }

      if (trdest != -1)
      {
         hscal_BI_MPI_Reduce_nocomm(
              ctxt, N, length, vvop, dest,
              nprow, npcol, myrow, mycol, tcomp, tcomm
         );
         if (ctxt->scp->Iam == dest)
	 {
	    hscal_BI_cvmcopy_nocomm(
                 Mpval(m), Mpval(n), tlda,
                 nprow, npcol, myrow, mycol, tcomp
            );
	    if (Mpval(ldia) != -1)
            {
               hscal_BI_TransDist_nocomm(
                    ctxt, tscope, Mpval(m), Mpval(n), rA, cA, tldia,
                    dist,
                    trdest, Mpval(cdest),
                    nprow, npcol, myrow, mycol, tcomp
               );
               free(dist);
            }
	 }
      }
      else
      {
         hscal_BI_MPI_Allreduce_nocomm(
              ctxt, N, length, vvop,
              nprow, npcol, myrow, mycol, tcomp, tcomm
         );

	 hscal_BI_cvmcopy_nocomm(
              Mpval(m), Mpval(n), tlda,
              nprow, npcol, myrow, mycol, tcomp
         );
         if (Mpval(ldia) != -1)
         {
            hscal_BI_TransDist_nocomm(
                 ctxt, tscope, Mpval(m), Mpval(n), rA, cA, tldia,
                 dist,
                 trdest, Mpval(cdest),
                 nprow, npcol, myrow, mycol, tcomp
            );
            free(dist);
         }
      }
      return;
      break;
   case 'i':
      hscal_BI_MringComb_nocomm(
           ctxt, N, length, vvop, dest, 1,
           nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 'd':
      hscal_BI_MringComb_nocomm(
           ctxt, N, length, vvop, dest, -1,
           nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 's':
      hscal_BI_MringComb_nocomm(
           ctxt, N, length, vvop, dest, 2,
           nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 'm':
      hscal_BI_MringComb_nocomm(
           ctxt, N, length, vvop, dest, ctxt->Nr_co,
           nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case '1':
   case '2':
   case '3':
   case '4':
   case '5':
   case '6':
   case '7':
   case '8':
   case '9':
      hscal_BI_TreeComb_nocomm(
           ctxt, N, length, vvop, dest, ttop-47,
           nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 'f':
      hscal_BI_TreeComb_nocomm(
           ctxt, N, length, vvop, dest, FULLCON,
           nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 't':
      hscal_BI_TreeComb_nocomm(
           ctxt, N, length, vvop, dest, ctxt->Nb_co,
           nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 'h':
/*
 *    Use bidirectional exchange if everyone wants answer
 */
      if ( (trdest == -1) && !(ctxt->TopsCohrnt) )
         hscal_BI_BeComb_nocomm(
              ctxt, N, length, vvop,
              nprow, npcol, myrow, mycol, tcomp, tcomm
         );
      else
         hscal_BI_TreeComb_nocomm(
              ctxt, N, length, vvop, dest, 2,
              nprow, npcol, myrow, mycol, tcomp, tcomm
         );
      break;
   default :
      BI_BlacsErr_nocomm(ctxt, __LINE__, __FILE__, "Unknown topology '%c'",
                  ttop);
   }

/*
 * If I am selected to receive answer
 */
   if ( (ctxt->scp->Iam == dest) || (dest == -1) )
   {
/*
 *    Translate the distances stored in the latter part of bp->Buff into
 *    process grid coordinates, and output these coordinates in the
 *    arrays rA and cA.
 */
      if (Mpval(ldia) != -1)
      {
         hscal_BI_TransDist_nocomm(
              ctxt, tscope, Mpval(m), Mpval(n), rA, cA, tldia,
              dist, trdest, Mpval(cdest),
              nprow, npcol, myrow, mycol, tcomp
         );
         free(dist);
      }
/*
 *    Unpack the amx array
 */
      if (contiguous == 1) 
         hscal_BI_cvmcopy_nocomm(
              Mpval(m), Mpval(n), tlda,
              nprow, npcol, myrow, mycol, tcomp
         );
   }
}

#endif
