
#ifndef __HSCAL_BLACS_CCGSUM2D_HSCALFUNC_HH
#define __HSCAL_BLACS_CCGSUM2D_HSCALFUNC_HH

#include "../../../include/scalapack.h"

void hscal_Ccgsum2d(
     HSCALBLACSCONTEXT *ctxt, char *scope, char *top, 
     int m, int n, 
     int lda, int rdest, int cdest,
     int nprow, int npcol, int myrow, int mycol, 
     double *tcomp, int *tcomm
)
/*
 *  -- V1.1 BLACS routine --
 *  University of Tennessee, May 1, 1996
 *  Written by Clint Whaley.
 *
 *  Purpose
 *  =======
 *  Combine sum operation for complex rectangular matrices.
 *
 *  Arguments
 *  =========
 *
 *  ConTxt  (input) Ptr to int
 *          Index into MyConTxts00 (my contexts array).
 *
 *  SCOPE   (input) Ptr to char
 *          Limit the scope of the operation.
 *          = 'R' :   Operation is performed by a process row.
 *          = 'C' :   Operation is performed by a process column.
 *          = 'A' :   Operation is performed by all processes in grid.
 *
 *  TOP     (input) Ptr to char
 *          Controls fashion in which messages flow within the operation.
 *
 *  M       (input) Ptr to int
 *          The number of rows of the matrix A.  M >= 0.
 *
 *  N       (input) Ptr to int
 *          The number of columns of the matrix A.  N >= 0.
 *
 *  A       (output) Ptr to complex two dimensional array
 *          The m by n matrix A.  Fortran77 (column-major) storage
 *          assumed.
 *
 *  LDA     (input) Ptr to int
 *          The leading dimension of the array A.  LDA >= M.
 *
 *  RDEST   (input) Ptr to int
 *          The process row of the destination of the sum.
 *          If rdest == -1, then result is left on all processes in scope.
 *
 *  CDEST   (input) Ptr to int
 *          The process column of the destination of the sum.
 *          If rdest == -1, then CDEST ignored.
 *
 * ------------------------------------------------------------------------
 */
{
/*
 *  Variable Declarations
 */
   char ttop, tscope;
   int N, lth, dest, tlda, trdest, ierr;
   int contiguous = 0;

 /*
  * Fill up the uninitialized fields
  */
   ctxt->rscp.Iam = mycol;
   ctxt->cscp.Iam = myrow;
   ctxt->pscp.Iam = ctxt->ascp.Iam = myrow*npcol+mycol;

   ttop = F2C_CharTrans(top);
   ttop = Mlowcase(ttop);
   tscope = F2C_CharTrans(scope);
   tscope = Mlowcase(tscope);
/*
 *  If the user has set the default combine topology, use it instead of
 *  BLACS default
 */
   if (ttop == ' ') ttop = DefCombTop;
   if (Mpval(cdest) == -1) trdest = -1;
   else trdest = Mpval(rdest);
   if (Mpval(lda) >= Mpval(m)) tlda = Mpval(lda);
   else tlda = Mpval(m);
   switch(tscope)
   {
   case 'r':
   {
      ctxt->scp = &ctxt->rscp;
      if (trdest == -1) dest = -1;
      else dest = Mpval(cdest);
      break;
   }
   case 'c':
   {
      ctxt->scp = &ctxt->cscp;
      dest = trdest;
      break;
   }
   case 'a':
   {
      ctxt->scp = &ctxt->ascp;
      if (trdest == -1) dest = -1;
      else dest = Mvkpnum(ctxt, trdest, Mpval(cdest));
      break;
   }
   default:
      BI_BlacsErr_nocomm(ctxt, __LINE__, __FILE__, "Unknown scope '%c'",
                  tscope);
      return;
   }


/*
 * It's not defined how MPI reacts to 0 element reductions, so use BLACS 1-tree
 * topology if we've got one.  Also, we can't use MPI functions if we need to
 * guarantee repeatability.
 */
   if (ttop == ' ')
      if ( (Mpval(m) < 1) || (Mpval(n) < 1) || (ctxt->TopsRepeat) ) ttop = '1';
   N = Mpval(m) * Mpval(n);
   lth = N * sizeof(SCOMPLEX);

/*
 * If A is contiguous, we can use it as one of the buffers
 */
   if ( (Mpval(m) == tlda) || (Mpval(n) == 1) )
   {
       contiguous = 1;
   }
/*
 * Otherwise, we must allocate both buffers
 */
   else
   {
      hscal_BI_cmvcopy_nocomm(
           Mpval(m), Mpval(n), tlda, 
           nprow, npcol, myrow, mycol, tcomp
      );
   }

   switch(ttop)
   {
   case ' ':         /* use MPI's reduction by default */
      if (dest != -1)
      {
         hscal_BI_MPI_Reduce_nocomm(
               ctxt, N, lth, hscal_BI_cMPI_sum, dest,
               nprow, npcol, myrow, mycol, tcomp, tcomm
         );

         if (ctxt->scp->Iam == dest)
         {
	    hscal_BI_cvmcopy_nocomm(
                 Mpval(m), Mpval(n), tlda, 
                 nprow, npcol, myrow, mycol, tcomp
            );
         }
      }
      else
      {
         hscal_BI_MPI_Allreduce_nocomm(
               ctxt, N, lth, hscal_BI_cMPI_sum,
               nprow, npcol, myrow, mycol, tcomp, tcomm
         );

         hscal_BI_cvmcopy_nocomm(
              Mpval(m), Mpval(n), tlda, 
              nprow, npcol, myrow, mycol, tcomp
         );
      }
      return;
      break;
   case 'i':
   {
      int nrings = 1;
      hscal_BI_MringComb_nocomm(
        ctxt, N, lth, hscal_BI_cvvsum, dest, nrings,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   }
   case 'd':
   {
      int nrings = -1;
      hscal_BI_MringComb_nocomm(
        ctxt, N, lth, hscal_BI_cvvsum, dest, nrings,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   }
   case 's':
   {
      int nrings = 2;
      hscal_BI_MringComb_nocomm(
        ctxt, N, lth, hscal_BI_cvvsum, dest, nrings,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   }
   case 'm':
      hscal_BI_MringComb_nocomm(
        ctxt, N, lth, hscal_BI_cvvsum, dest, ctxt->Nr_co,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case '1':
   case '2':
   case '3':
   case '4':
   case '5':
   case '6':
   case '7':
   case '8':
   case '9':
   {
      int nbranches = ttop-47;
      hscal_BI_TreeComb_nocomm(
        ctxt, N, lth, hscal_BI_cvvsum, dest, nbranches,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   }
   case 'f':
   {
      int nbranches = FULLCON;
      hscal_BI_TreeComb_nocomm(
        ctxt, N, lth, hscal_BI_cvvsum, dest, nbranches,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   }
   case 't':
      hscal_BI_TreeComb_nocomm(
        ctxt, N, lth, hscal_BI_cvvsum, dest, ctxt->Nb_co,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 'h':
/*
 *    Use bidirectional exchange if everyone wants answer
 */
      if ( (trdest == -1) && !(ctxt->TopsCohrnt) )
      {
         hscal_BI_BeComb_nocomm(
           ctxt, N, lth, hscal_BI_cvvsum,
           nprow, npcol, myrow, mycol, tcomp, tcomm
         );
      }
      else
      {
         int nbranches = 2;
         hscal_BI_TreeComb_nocomm(
           ctxt, N, lth, hscal_BI_cvvsum, dest, nbranches,
           nprow, npcol, myrow, mycol, tcomp, tcomm
         );
      }
      break;
   default :
      BI_BlacsErr_nocomm(ctxt, __LINE__, __FILE__, "Unknown topology '%c'",
                  ttop);
   }

/*
 * If I am selected to receive answer
 */
   if (contiguous == 1)
      if ( (ctxt->scp->Iam == dest) || (dest == -1) )
         hscal_BI_cvmcopy_nocomm(
              Mpval(m), Mpval(n), tlda,
              nprow, npcol, myrow, mycol, tcomp
         );
}

#endif
