
#ifndef __HSCAL_BLACS_CDGEBS2D_MACRO_HH
#define __HSCAL_BLACS_CDGEBS2D_MACRO_HH

#include "../../../include/scalapack.h"

void hscal_Cdgebs2d(
     HSCALBLACSCONTEXT *ctxt, char *scope, char *top, 
     int m, int n, int lda,
     int nprow, int npcol, int myrow, int mycol, 
     double *tcomp, int *tcomm
)
/*
 *  -- V1.1 BLACS routine --
 *  University of Tennessee, May 1, 1996
 *  Written by Clint Whaley.
 *
 *  Purpose
 *  =======
 *  Broadcast/send for general double precision arrays.
 *
 *  Arguments
 *  =========
 *
 *  ConTxt  (input) Ptr to int
 *          Index into MyConTxts00 (my contexts array).
 *
 *  SCOPE   (input) Ptr to char
 *          Limit the scope of the operation.
 *          = 'R' :   Operation is performed by a process row.
 *          = 'C' :   Operation is performed by a process column.
 *          = 'A' :   Operation is performed by all processes in grid.
 *
 *  TOP     (input) Ptr to char
 *          Controls fashion in which messages flow within the operation.
 *
 *  M       (input) Ptr to int
 *          The number of rows of the matrix A.  M >= 0.
 *
 *  N       (input) Ptr to int
 *          The number of columns of the matrix A.  N >= 0.
 *
 *  A       (input) Ptr to double precision two dimensional array
 *          The m by n matrix A.  Fortran77 (column-major) storage
 *          assumed.
 *
 *  LDA     (input) Ptr to int
 *          The leading dimension of the array A.  LDA >= M.
 *
 * ------------------------------------------------------------------------
 */
{
   char ttop, tscope;
   int error, tlda;
   int N, lth;

 /*
  * Fill up the uninitialized fields
  */
   ctxt->rscp.Iam = mycol;
   ctxt->cscp.Iam = myrow;
   ctxt->pscp.Iam = ctxt->ascp.Iam = myrow*npcol+mycol;

/*
 * get context, lowcase char variables, and perform parameter checking
 */
   ttop = F2C_CharTrans(top);
   ttop = Mlowcase(ttop);
   tscope = F2C_CharTrans(scope);
   tscope = Mlowcase(tscope);
/*
 *  If the user has set the default broadcast topology, use it instead of
 *  BLACS default
 */
   if (ttop == ' ') ttop = DefBSTop;
   if (Mpval(lda) < Mpval(m)) tlda = Mpval(m);
   else tlda = Mpval(lda);

   switch(tscope)
   {
   case 'r':
      ctxt->scp = &ctxt->rscp;
      break;
   case 'c':
      ctxt->scp = &ctxt->cscp;
      break;
   case 'a':
      ctxt->scp = &ctxt->ascp;
      break;
   default:
      BI_BlacsErr_nocomm(ctxt, __LINE__, __FILE__, "Unknown scope '%c'",
                  tscope);
      return;
   }

   N = Mpval(m) * Mpval(n);
   lth = N * sizeof(double);

/*
 * Call correct topology for BS/BR
 */
   switch(ttop)
   {
   case ' ':         /* If using default topology, use MPI native broadcast */
      hscal_BI_MPI_Bcast_nocomm(
        ctxt, N, lth, ctxt->scp->Iam,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 'h':
   {
      hscal_BI_HypBS_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm,
        &error,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );

      if (error == NPOW2) 
      {
         hscal_BI_TreeBS_nocomm(
                            ctxt, N, lth, hscal_BI_MPI_Send_nocomm, 2,
                            nprow, npcol, myrow, mycol, tcomp, tcomm
                          );
      }

      break;
   }
   case '1':
   case '2':
   case '3':
   case '4':
   case '5':
   case '6':
   case '7':
   case '8':
   case '9':
      hscal_BI_TreeBS_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm, ttop-47,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 't':
      hscal_BI_TreeBS_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm, ctxt->Nb_bs,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 'i':
      hscal_BI_IdringBS_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm, 1,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 'd':
      hscal_BI_IdringBS_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm, -1,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 's':
      hscal_BI_SringBS_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 'f':
   {
      int npaths = FULLCON;
      hscal_BI_MpathBS_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm, npaths,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   }
   case 'm':
      hscal_BI_MpathBS_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm, ctxt->Nr_bs,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   default :
      BI_BlacsErr_nocomm(ctxt, __LINE__, __FILE__, "Unknown topology '%c'",
                  ttop);
      return;
   }
}

#endif
