
#ifndef __HSCAL_BLACS_CSGEBR2D_HSCALFUNC_HH
#define __HSCAL_BLACS_CSGEBR2D_HSCALFUNC_HH

#include "../../../include/scalapack.h"

void hscal_Csgebr2d(
     HSCALBLACSCONTEXT *ctxt, char *scope, char *top, int m, int n,
     int lda, int rsrc, int csrc,
     int nprow, int npcol, int myrow, int mycol, 
     double *tcomp, int *tcomm 
)
/*
 *  -- V1.1 BLACS routine --
 *  University of Tennessee, May 1, 1996
 *  Written by Clint Whaley.
 *
 *  Purpose
 *  =======
 *  Broadcast/receive for general real arrays.
 *
 *  Arguments
 *  =========
 *
 *  ConTxt  (input) Ptr to int
 *          Index into MyConTxts00 (my contexts array).
 *
 *  SCOPE   (input) Ptr to char
 *          Limit the scope of the operation.
 *          = 'R' :   Operation is performed by a process row.
 *          = 'C' :   Operation is performed by a process column.
 *          = 'A' :   Operation is performed by all processes in grid.
 *
 *  TOP     (input) Ptr to char
 *          Controls fashion in which messages flow within the operation.
 *
 *  M       (input) Ptr to int
 *          The number of rows of the matrix A.  M >= 0.
 *
 *  N       (input) Ptr to int
 *          The number of columns of the matrix A.  N >= 0.
 *
 *  A       (output) Ptr to real two dimensional array
 *          The m by n matrix A.  Fortran77 (column-major) storage
 *          assumed.
 *
 *  LDA     (input) Ptr to int
 *          The leading dimension of the array A.  LDA >= M.
 *
 *
 *  RSRC    (input) Ptr to int
 *          The process row of the source of the matrix.
 *
 *  CSRC    (input) Ptr to int
 *          The process column of the source of the matrix.
 *
 *
 * ------------------------------------------------------------------------
 */
{
   int N, lth, src, tlda, error, one=1;
   char ttop, tscope;

 /*
  * Fill up the uninitialized fields
  */
   ctxt->rscp.Iam = mycol;
   ctxt->cscp.Iam = myrow;
   ctxt->pscp.Iam = ctxt->ascp.Iam = myrow*npcol+mycol;

   ttop = F2C_CharTrans(top);
   ttop = Mlowcase(ttop);
   tscope = F2C_CharTrans(scope);
   tscope = Mlowcase(tscope);

/*
 *  If the user has set the default broadcast topology, use it instead of
 *  BLACS default
 */
   if (ttop == ' ') ttop = DefBSTop;
   if (Mpval(m) <= Mpval(lda)) tlda = Mpval(lda);
   else tlda = Mpval(m);

   switch(tscope)
   {
   case 'r':
      ctxt->scp = &ctxt->rscp;
      src = Mpval(csrc);
      break;
   case 'c':
      ctxt->scp = &ctxt->cscp;
      src = Mpval(rsrc);
      break;
   case 'a':
      ctxt->scp = &ctxt->ascp;
      src = Mvkpnum(ctxt, Mpval(rsrc), Mpval(csrc));
      break;
   default:
      BI_BlacsErr_nocomm(ctxt, __LINE__, __FILE__, "Unknown scope '%c'",
                  tscope);
      return;
   }

   N = Mpval(m) * Mpval(n);
   lth = N * sizeof(float);

   switch(ttop)
   {
   case ' ':         /* If using default topology, use MPI native broadcast */
      hscal_BI_MPI_Bcast_nocomm(
        ctxt, N, lth, src,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 'h':
   {
      hscal_BI_HypBR_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm, src,
        &error,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      if (error == NPOW2) 
      {
          hscal_BI_TreeBR_nocomm(
                            ctxt, N, lth, hscal_BI_MPI_Send_nocomm, src, 2,
                            nprow, npcol, myrow, mycol, tcomp, tcomm
                          );
      }
      break;
   }
   case '1':
   case '2':
   case '3':
   case '4':
   case '5':
   case '6':
   case '7':
   case '8':
   case '9':
      hscal_BI_TreeBR_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm, src, ttop-47,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 't':
      hscal_BI_TreeBR_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm, src, ctxt->Nb_bs,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 'i':
      hscal_BI_IdringBR_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm, src, 1,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 'd':
      hscal_BI_IdringBR_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm, src, -1,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 's':
      hscal_BI_SringBR_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm, src,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 'm':
      hscal_BI_MpathBR_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm, src, ctxt->Nr_bs,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 'f':
   {
      int npaths = FULLCON;
      hscal_BI_MpathBR_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm, src, npaths,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   }
   default :
      BI_BlacsErr_nocomm(ctxt, __LINE__, __FILE__, "Unknown topology '%c'",
                  ttop);
      return;
   }
}

#endif

