
#include "be_print.h"
#include "netstuff.h"

/*
 * Prints the net/subnet predicate (boolean expression, having
 * value 1 for processes inside the network and 0 in the either
 * case).
 *
 * @param inNet net or subnet
 */
static void print_net_predicate(tTree inNet)
{
    print2file("", "(");
    if(inNet->Kind == kMPC_Subnet) {
        print_net_predicate(inNet->MPC_Subnet.Distribution);
        print2file("", "&&(");
        print_sexpr(inNet->MPC_Subnet.Predicate);
        print2file("", ")");
    } else {
        print2file("", "%s(&%s)", MEMBER, VarName(inNet));
    }
    print2file("", ")"); 
    
}

/* Traverses declarations and calls inF for every
 * net or subnet.
 * 
 * @param inDecls declrations list
 * @param inF function to call in each node
 */
static void traverse_nets_and_subnets(tTree inDecls, void inF(tTree, tTree))
{
    tTree p, pNet;
    
    for(p = inDecls; p->Kind != kMPC_FreeNode; p = p->MPC_Decls.Next) {
        switch (p->Kind) {
        case kMPC_NetDecl:           
            for(pNet = p->MPC_NetDecl.Net;
                pNet->Kind != kMPC_FreeNode;
                pNet = pNet->MPC_Net.Next){
                inF(p, pNet);
            }            
            break;
        case kMPC_SubnetDecl:            
            for(pNet = p->MPC_SubnetDecl.Subnet;            
                pNet->Kind != kMPC_FreeNode;
                pNet = pNet->MPC_Subnet.Next){
                inF(p, pNet);
            }           
            break;
        }
    }
}

/* Traverses declarations in the reverse order and calls inF for every
 * net or subnet.
 * 
 * @param inDecls declrations list
 * @param inF function to call in each node
 */
static void reverse_nets_and_subnets(tTree inEnclosingStat, tTree inDecls, void inF(tTree, tTree, tTree))
{
    tTree p, q, pNet;
    
    for(q = p = inDecls; p->Kind != kMPC_FreeNode; q = p, p = p->MPC_Decls.Next);
    
    for(; q->Kind != kMPC_FreeNode; q = q->MPC_Decls.Prev)
     {
        switch (q->Kind) {
        case kMPC_NetDecl:           
            for(pNet = p = q->MPC_NetDecl.Net;
                p->Kind != kMPC_FreeNode;
                pNet = p, p = p->MPC_Net.Next);
                
            for(;
                pNet->Kind != kMPC_FreeNode;
                pNet = pNet->MPC_Net.Prev){
                inF(inEnclosingStat, q, pNet);
            }            
            break;
        case kMPC_SubnetDecl:            
            for(pNet = p = q->MPC_SubnetDecl.Subnet;            
                p->Kind != kMPC_FreeNode;
                pNet = p, p = p->MPC_Subnet.Next);
                
            for(;            
                pNet->Kind != kMPC_FreeNode;
                pNet = pNet->MPC_Subnet.Prev){
                inF(inEnclosingStat, q, pNet);
            }           
            break;
        }
    }
}


/**
 * Prints subnet creation. 
 *
 * @param inDecl subnet declaration
 *
 * @param inIsFlexible true if subnetwork is flexible
 *
 */
void print_create_subnet(tTree inDecl){
    c_blank;
    print2file("", "%s(%s, ", MPC_SUBNET_CREATE, VarName(inDecl->MPC_Subnet.Distribution));
    print2file("", "%s, ", VarName(inDecl));
    print_sexpr(inDecl->MPC_Subnet.Predicate);
    print2file("", ", %d)\n", inDecl->MPC_Subnet.Flexible);
}


/*
 * Prints topology parameters before net creation.
 *
 * @param inDecl network declaraion
 */
static void print_fill_topo_params(tTree inDecl)
{
    tTree pNetTypeSpec = inDecl->MPC_Net.NetType;
    tTree pNetType, pArgList, pParamList;
    tTree old_pCurParamList = pCurParamList;
    tTree old_pCurExprs = pCurExprs;
    
    {
        tTree pNext =  pNetTypeSpec-> MPC_NetTypeSpecifier.NetType;
        while (pNext->Kind == kMPC_NetTypeSpecifier) {
            pNetTypeSpec = pNext;
            pNext = pNext->MPC_NetTypeSpecifier.NetType;            
        }
    }
    pNetType = pNetTypeSpec->MPC_NetTypeSpecifier.NetType;
    PointerControl(pNetType);
    pParamList = pNetType->MPC_NetType.ParamList;
    PointerControl(pParamList);
    pCurParamList = pParamList;
    pArgList =pNetTypeSpec->MPC_NetTypeSpecifier.ArgList;
    PointerControl(pArgList);
    pCurExprs = pArgList;
    if (pParamList->Kind != kMPC_FreeNode) {
        c_blank;
        print2file("", "%s.count=", VarName(inDecl));
        print_ParamNumber(pArgList, -1);
        print2file("", ";\n");
        print_CopyParams(inDecl, pNetType, pArgList);
    }
    pCurParamList = old_pCurParamList;
    pCurExprs = old_pCurExprs;
}

/*
 * Prints net creation.
 *
 * @param inDecl net declaration
 */ 
static void print_create_net(tTree inDecl)
{
    print_set_position(inDecl->MPC_Net.Pos);    
    c_blank;
    print2file("","if(");
    print_net_predicate(inDecl->MPC_Net.Distribution);
    print2file("",")");
    BRACE_OPEN;
    
    if(inDecl->MPC_Net.Topology != CONST){
        print_fill_topo_params(inDecl);
    }
    
    c_blank;
    print2file("", "%s(%d, &%s, 1);\n", 
        MPC_NET_CREATE_GLOBAL,      
        inDecl->MPC_Net.UniqueNumber, 
        VarName(inDecl));
    BRACE_CLOSE;
    c_blank;
    print2file("","else");
    BRACE_OPEN;
    c_blank;
    print2file("", "%s(%d, &%s, 0);\n", 
        MPC_NET_CREATE_GLOBAL,      
        inDecl->MPC_Net.UniqueNumber, 
        VarName(inDecl));
    BRACE_CLOSE;
}

/*
 * Prints net and subnet creation.
 *
 * @param inNetDecl declaration
 * @param inNet net/subnet
 */
static void print_create_net_or_subnet(tTree inNetDecl, tTree inNet)
{
    switch(inNetDecl->Kind){
    case kMPC_NetDecl:
        print_create_net(inNet);
    break;
    case kMPC_SubnetDecl:        
        print_create_subnet(inNet);
    break;
    }
}
 

 
 
/*
 * Traverses declarations and prints found nets and subnets
 * creation.
 *
 * @param inDecls declration list
 */
void print_create_nets_and_subnets(tTree inDecls)
{
    if(inDecls != NoTree)
        traverse_nets_and_subnets(inDecls, print_create_net_or_subnet);
}

/*
 * Prints net destroyment.
 *
 * @param inDecl net declaration
 */
static void print_free_net(tTree inEnclosingStat, tTree inDecl, tTree inNet)
{
    if((inDecl == NoTree) || !(inDecl->MPC_NetDecl.Flag.Static || inDecl->MPC_NetDecl.Flag.Extern)) {
        print_set_position(inEnclosingStat->MPC_Compound.EndPos);
        c_blank;
        print2file("", "%s(&%s);\n", 
            MPC_NET_FREE_GLOBAL,              
            VarName(inNet));
    }
}    
 
/*
 * Prints subnet destroyment.
 *
 * @param inDecl net declaration
 */
void print_free_subnet(tTree inDecl)
{
    c_blank;
    print2file("","if");
    print_net_predicate(inDecl->MPC_Subnet.Distribution);
    BRACE_OPEN;
    c_blank;
    print2file("", "%s(&%s);\n", 
        MPC_NET_FREE_GLOBAL,              
        VarName(inDecl));
    BRACE_CLOSE;        
}   
 
/*
 * Prints net and subnet destroyment.
 *
 * @param inDecl declaration
 */
static void print_free_net_or_subnet(tTree inEnclosingStat, tTree inNetDecl, tTree inNet)
{    
    switch(inNetDecl->Kind){
    case kMPC_NetDecl:
        print_free_net(inEnclosingStat, inNetDecl, inNet);
    break;
    case kMPC_SubnetDecl:        
        print_free_subnet(inNet);
    break;
    }
}

/*
 * Traverses declarations and prints found nets and subnet
 * destroyment.
 *
 * @param inDecls declaration list
 */
 void print_free_nets_and_subnets(tTree inEnclosingStat, tTree inDecls)
 {
    if(inDecls != NoTree)
        reverse_nets_and_subnets(inEnclosingStat, inDecls, print_free_net_or_subnet);
        
 }
 
/*
 * Returns the statement above.
 *
 * @param inStat current statement
 *
static tTree get_up_stat(tTree inStat)
{
    for(;inStat->Kind != kMPC_FreeNode; inStat = inStat->MPC_Stat.Next);
    return inStat->MPC_FreeNode.Parent;
}
*/

/*
 * Prints nets destroyment for all created nets in the function.
 * 
 * @param inStat current statement
 */
 void print_global_free_nets_and_subnets(tTree inStat)
 {
    for(;inStat->Kind != kMPC_Function; inStat = get_parent_stat(inStat))
        if(inStat->Kind == kMPC_Compound)            
            print_free_nets_and_subnets(inStat, inStat->MPC_Compound.Decls);                
        
    
    if(inStat->MPC_Function.Kind == NETWORK){
        if(inStat->MPC_Function.Type->MPC_FunctionType.NetworkParamList->Kind != kMPC_FreeNode)
            print_free_net(inStat->MPC_Function.Stats, NoTree, inStat->MPC_Function.Type->MPC_FunctionType.NetParam);        
    }
 }
