
/*****************************************************************************/
/*                            Revision history                               */
/*                                                                           */
/* DD-MM-YYYY          Description (Developer Name)                          */
/* ----------          ------------------------------------------------------*/
/* 01-11-2007  --      Consolidation of the MPC CVS rts and MPC3XX rts       */
/*                     into a single rts (Ravi Reddy)                        */
/* 12-11-2007  --      Creation of the topology environment and saving it    */
/*                     to a file if provided by the user as an input         */
/*                     (Ravi Reddy)                                          */
/*****************************************************************************/

#include "mpCinternal.h"
#include "mpCtopo.h"
#include <stdlib.h>
#include <stdio.h>
#include <float.h>
#include <ctype.h>
#include <sys/times.h>
#include <limits.h>
#include <unistd.h>
#include <sys/utsname.h>
#include <string.h>
extern long sysconf(int);

#include <mpi.h>

/************************ Global Structures & Variables ******************************************/
/*
* Number of iterations of the benchmark code
*/
#define       MPC_TOPO_ITERS           200

/*
* Number of repeated testings of communication links
*/
#define       MPC_TOPO_REPS            3

/*
* Number of attributes: speed, memory, scalability
*/
#define       MPC_TOPO_NUM_ATTRTYPES   3

typedef struct {
        double speed; 
        double memory; 
        double scalability;
} MPC_tNodeAttribute_i;

/************************ MPC_SeqMult_i *********************************************************/

int MPC_SeqMult_i(double *x, double *y, double *z, int n)
{
  int i, j, k, ixn;
  double s;

  for (i = 0; i < n; i++)
  {
      ixn = i*n;
      for (j = 0; j < n; j++)
      {
          for (k = 0, s = 0.0; k < n; k++)
          {
            s += x[ixn+k] * (double)(y[k*n+j]);
          }
          z[ixn+j] = s;
      }
  }

  return MPC_OK;
}

/************************ MPC_Initialize_input_i ************************************************/

int MPC_Initialize_input_i(double **x, double **y, int n)
{
  int i, j, ixn;

  *x = (double*)calloc(n*n, sizeof(double));

  if ((*x) == NULL)
  {
     return MPC_ERR_NOMEM;
  }

  *y = (double*)calloc(n*n, sizeof(double));

  if ((*y) == NULL)
  {
     return MPC_ERR_NOMEM;
  }

  for (i = 0; i < n; i++)
  {
      ixn = i*n;
      for (j = 0; j < n; j++)
      {
        if (i == j)
        {
           (*x)[ixn+j] = 2.0;
           (*y)[ixn+j] = 3.0;
        }
        else
        {
           (*x)[ixn+j] = 0.001;
           (*y)[ixn+j] = 0.001;
        }
      }
  }

  return MPC_OK;
}

/******************************* MPC_Actual_num_i ***********************************************/

int MPC_Actual_num_i(
    int N,
    char (*names)[MPC_MAX_PROCESSOR_NAME], 
    int **actual_ranks, 
    int **v2a
)
{
  int n, i, j, k;
  char **pname, **pi, **pj;

  n = N;

  pname = (char**)calloc(N, sizeof(*pname));
  
  if (pname == NULL)
  {
     return -1;
  }

  *v2a = (int*)calloc(N,sizeof(int));

  if (*v2a == NULL)
  {
     return -1;
  }

  *actual_ranks = (int*)calloc(N, sizeof(int));

  if (*actual_ranks == NULL)
  {
     return -1;
  }

  for (i = 0; i < N; i++) 
  {
    pname[i] = names[i];
    (*v2a)[i] = -1;
  }

  for (pi=pname, i=0, k=0; i < N; pi++, i++) 
  {
      if (*pi != NULL)
      {
         (*actual_ranks)[k++]=i;
      }
      else
      {
         continue;
      }

      for (pj = pi+1, j = i+1; j < N; pj++, j++) 
      {
        if (*pj == NULL)
        {
           continue;
        }

        if (strcmp(*pj, *pi) == 0) 
        {
           n--;
           (*v2a)[j] = i;
           *pj = NULL;
        }
      }
  }

  free(pname);

  return n;
}

/*********************************** MPC_Nearest_i ***********************************************/

int MPC_Nearest_i(double x)
{
  int low, hi;
 
  low=(int)x;
  hi=low+1;
  if (x-low < hi-x)
  {
     return low;
  }
  else
  {
     return hi;
  }

  return -1;
}

/************************ MPC_Initialize_topology_environment ************************************/

Int MPC_Initialize_topology_environment(
       Int save_to_topofile,
       char *wtopofile,
       Int* penv_length,
       Int** penvironment
)
{
    /* bobr: removed from arg list */
    MPI_Comm mpi_comm_world = MPI_COMM_WORLD;

    /*
    * My node name
    * List of nodenames
    */
    char mynodename[MPC_MAX_PROCESSOR_NAME];
    char (*nodenames)[MPC_MAX_PROCESSOR_NAME];

    /*
    * Return code from function calls
    * My rank
    * The size of the communication universe
    * The length of the name of my processor
    * The rank of the host process
    */
    int rc, me, commworldsize, plen, host = 0;
    static int *basic_ranks, *virtual2basic;

    /*
    * Data for derived type
    */
    static int blocklengths[MPC_TOPO_NUM_ATTRTYPES] = {1, 1, 1};
    static MPI_Aint displacements[MPC_TOPO_NUM_ATTRTYPES];
    static MPI_Datatype types[MPC_TOPO_NUM_ATTRTYPES], attrtype;

    /*
    * Hierarchy of clusters --- Although this code only supports one cluster
    */
    int (*cluster_stack)[2], cluster_pointer, cluster_num, cluster_process_num;

    /*
    * Number of unique nodes
    * Number of processes per node
    */
    int number_of_unique_nodes, *number_of_processes_per_node;

    /*
    * Flag indicating if the topo file can be opened for writing
    */
    int fileerr = 0;

    /*
    * Link count
    */
    int num_of_link_nodes, lncount = 0;

    /*
    * The total number of processes --- must be equal to comm world size
    */
    int total_processes = 0;

    /*
    * Text file to write the topology information
    */
    FILE *wtopofp;

    /*
    * Attributes of a node --- speed, memory, scalability
    */
    MPC_tNodeAttribute_i my_attribute, *processesAttributes, *processorsAttributes;
    
    /*
    * Group data
    * MPI group and communicator containing a process from each node
    */
    MPI_Group mpi_group_world, mpi_group_of_nodes;
    MPI_Comm mpi_comm_of_nodes;

    /*
    * The name of my node
    */
    rc = MPI_Get_processor_name(
            mynodename,
            &plen
    );

    rc = I2C(rc);

    if (rc != MPC_OK)
    {
       return rc;
    }

    rc = MPI_Comm_group(mpi_comm_world, &mpi_group_world);

    rc = I2C(rc);

    if (rc != MPC_OK)
    {
       return rc;
    }

    rc = MPI_Comm_size(mpi_comm_world, &commworldsize);

    rc = I2C(rc);

    if (rc != MPC_OK)
    {
       return rc;
    }

    rc = MPI_Comm_rank(mpi_comm_world, &me);

    rc = I2C(rc);

    if (rc != MPC_OK)
    {
       return rc;
    }

    types[0] = MPI_DOUBLE;
    types[1] = MPI_DOUBLE;
    types[2] = MPI_DOUBLE;

    displacements[1] = (char*)&my_attribute.memory
                       -
                       (char*)&my_attribute.speed;
    displacements[2] = 2*displacements[1];

    rc = MPI_Type_struct(
            MPC_TOPO_NUM_ATTRTYPES,
            blocklengths, 
            displacements, 
            types, 
            &attrtype
    );

    rc = I2C(rc);

    if (rc != MPC_OK)
    {
       return rc;
    }

    rc = MPI_Type_commit(&attrtype);

    rc = I2C(rc);

    if (rc != MPC_OK)
    {
       return rc;
    }

    if (me == host)
    {
       nodenames = (char(*)[MPC_MAX_PROCESSOR_NAME])calloc(
                                                    commworldsize, 
                                                    MPC_MAX_PROCESSOR_NAME*sizeof(char)
       );

       if (nodenames == NULL)
       {
          return MPC_ERR_NOMEM;
       }
    }
       
    rc = MPI_Gather(
            &mynodename, 
            MPC_MAX_PROCESSOR_NAME, 
            MPI_CHAR, 
            nodenames,
            MPC_MAX_PROCESSOR_NAME,
            MPI_CHAR, 
            host,
            mpi_comm_world
    );

    rc = I2C(rc);

    if (rc != MPC_OK)
    {
       return rc;
    }

    if (me != host)
    {
       nodenames = (char(*)[MPC_MAX_PROCESSOR_NAME])calloc(
                                                    commworldsize, 
                                                    MPC_MAX_PROCESSOR_NAME*sizeof(char)
       );

       if (nodenames == NULL)
       {
          return MPC_ERR_NOMEM;
       }
    }
    
    if (me == host)
    {
       processesAttributes = (MPC_tNodeAttribute_i*)calloc(
                                                    commworldsize, 
                                                    sizeof(MPC_tNodeAttribute_i)
       );

       if (processesAttributes == NULL)
       {
          return MPC_ERR_NOMEM;
       }
    }

    rc = MPI_Bcast(
            nodenames, 
            commworldsize*MPC_MAX_PROCESSOR_NAME, 
            MPI_CHAR, 
            host,
            mpi_comm_world
    );

    rc = I2C(rc);

    if (rc != MPC_OK)
    {
       return rc;
    }

    /*
    * If Iam the host
    */
    if (me == host)
    {
       int i;

       number_of_unique_nodes = MPC_Actual_num_i(
                                   commworldsize,
                                   nodenames, 
                                   &basic_ranks, 
                                   &virtual2basic
       );

       if (number_of_unique_nodes == -1)
       {
          return MPC_ERR_NOMEM;
       }

       if (MPC_DEBUG > 1)
       {
          MPC_Debug_printf("host: Basic ranks are ");
          for (i = 0; i < commworldsize; i++)
          {
             MPC_Debug_printf("%d ", basic_ranks[i]);
          }
          MPC_Debug_printf("\n");

          MPC_Debug_printf("host: Virtual to abstract ranks are ");
          for (i = 0; i < commworldsize; i++)
          {
             MPC_Debug_printf("%d ", virtual2basic[i]);
          }
          MPC_Debug_printf("\n");
       }

       /*
       * Number of groups will be equal to number of unique nodes
       */
       ngroups = number_of_unique_nodes;
 
       /*
       * Fill in the topology information
       */
       groups = (MPC_Group_descr*)malloc(
                                  sizeof(MPC_Group_descr)
                                  *
                                  ngroups
       );

       if (groups == NULL)
       {
          return MPC_ERR_NOMEM;
       }

       /*
       * Only one cluster is assumed
       */
       num_of_link_nodes = ngroups + 1;

       link_nodes = (MPC_Hi_descr*)malloc(
                                   sizeof(MPC_Hi_descr)
                                   *
                                   num_of_link_nodes
       );

       if (link_nodes == NULL)
       {
          return MPC_ERR_NOMEM;
       }

       /*
       * Initialize the environment
       * This should be removed in the future
       */
       *penv_length = 1;

       *penvironment = (Int*)malloc(
                             sizeof(Int)
                             *
                             (*penv_length)
       );

       if (*penvironment == NULL)
       {
           return MPC_ERR_NOMEM;
       }

       (*penvironment)[0] = 0;
    }

    rc = MPI_Bcast(
            &number_of_unique_nodes, 
            1, 
            MPI_INT, 
            host, 
            mpi_comm_world
    ) ;

    rc = I2C(rc);

    if (rc != MPC_OK)
    {
       return rc;
    }

    if (me != host)
    {
       basic_ranks = (int*)calloc(
                           number_of_unique_nodes, 
                           sizeof(int)
       );

       if (basic_ranks == NULL)
       {
          return MPC_ERR_NOMEM;
       }
    }

    rc = MPI_Bcast(
            basic_ranks, 
            number_of_unique_nodes, 
            MPI_INT, 
            host, 
            mpi_comm_world
    );

    rc = I2C(rc);

    if (rc != MPC_OK)
    {
       return rc;
    }

    rc = MPI_Group_incl(
            mpi_group_world, 
            number_of_unique_nodes, 
            basic_ranks,
            &mpi_group_of_nodes
    );

    rc = I2C(rc);

    if (rc != MPC_OK)
    {
       return rc;
    }

    rc = MPI_Comm_create(
            mpi_comm_world, 
            mpi_group_of_nodes, 
            &mpi_comm_of_nodes
    );

    rc = I2C(rc);

    if (rc != MPC_OK)
    {
       return rc;
    }

    if (mpi_comm_of_nodes != MPI_COMM_NULL)
    {
       /*
       * If Iam the host
       */
       if (me == host)
       {
          if (save_to_topofile == 1)
          {
             wtopofp = fopen(wtopofile, "w");

             if (wtopofp == NULL)
             {
                fileerr = 1;
             }
          }
       }
    }

    rc = MPI_Bcast(
            &fileerr, 
            1, 
            MPI_INT, 
            host,
            mpi_comm_world
    );

    rc = I2C(rc);

    if (rc != MPC_OK)
    {
       return rc;
    }

    if (fileerr == 1)
    {
       return MPC_ERR_FILE;
    }

    {
       int i, j;

       double *a, *b, *c, diff, start1, start2, end1, end2;

       struct tms s, e;
 
       rc = MPC_Initialize_input_i(
               &a, 
               &b, 
               MPC_TOPO_ITERS
       );

       rc = I2C(rc);

       if (rc != MPC_OK)
       {
          return rc;
       }

       c = (double*)calloc(
                    MPC_TOPO_ITERS*MPC_TOPO_ITERS, 
                    sizeof(double)
       );
 
       if (c == NULL)
       {
          return MPC_ERR_NOMEM;
       }

       rc = MPI_Barrier(mpi_comm_world);

       rc = I2C(rc);

       if (rc != MPC_OK)
       {
          return rc;
       }

       if (mpi_comm_of_nodes != MPI_COMM_NULL) 
       {
          start1 = MPI_Wtime();
       }

       times(&s);
       MPC_SeqMult_i(a, b, c,  MPC_TOPO_ITERS);
       times(&e);

       if (mpi_comm_of_nodes != MPI_COMM_NULL) 
       {
          end1 = MPI_Wtime();
       }

       rc = MPI_Barrier(mpi_comm_world);
 
       rc = I2C(rc);

       if (rc != MPC_OK)
       {
          return rc;
       }

       diff = e.tms_utime - s.tms_utime;
 
       rc = MPI_Barrier(mpi_comm_world);

       rc = I2C(rc);

       if (rc != MPC_OK)
       {
          return rc;
       }

       if (mpi_comm_of_nodes != MPI_COMM_NULL)
       {
          start2 = MPI_Wtime();
          MPC_SeqMult_i(a, b, c, MPC_TOPO_ITERS);
          end2 = MPI_Wtime();
       }

       rc = MPI_Barrier(mpi_comm_world);

       rc = I2C(rc);

       if (rc != MPC_OK)
       {
          return rc;
       }

       free(a);
       free(b);
       free(c);

       if (mpi_comm_of_nodes != MPI_COMM_NULL)
       {
          my_attribute.scalability = (end2-start2)/(end1-start1);
       }

       /*
       * sysconf must be removed by number of floating point operations
       */
       my_attribute.speed = sysconf(_SC_CLK_TCK)/diff*MPC_POWER_COEFF;

       /*
       * No support for memory at the moment
       */
       my_attribute.memory = 0;

       rc = MPI_Gather(
               &my_attribute, 
               1, 
               attrtype, 
               processesAttributes, 
               1,
               attrtype, 
               host,
               mpi_comm_world
       );

       rc = I2C(rc);

       if (rc != MPC_OK)
       {
          return rc;
       }

       if (MPC_DEBUG > 1)
       {
          MPC_Debug_printf("me=%d: benchmark executed\n", me);
       }

       /*
       * If Iam the host
       */
       if (me == host)
       {
          int i, j, k, probes;

          processorsAttributes = (MPC_tNodeAttribute_i*)calloc(
                                                        number_of_unique_nodes, 
                                                        sizeof(MPC_tNodeAttribute_i)
          );

          if (processorsAttributes == NULL)
          {
             return MPC_ERR_NOMEM;
          }

          number_of_processes_per_node = (int*)calloc(
                                               number_of_unique_nodes, 
                                               sizeof(int)
          );

          if (number_of_processes_per_node == NULL)
          {
             return MPC_ERR_NOMEM;
          }

          if (MPC_DEBUG > 1)
          {
             MPC_Debug_printf("host: Number of unique nodes %d\n", number_of_unique_nodes);
          }

          for (i = 0; i < number_of_unique_nodes; i++) 
          {
              k = basic_ranks[i];

              for (j = k+1, probes = 1; j < commworldsize; j++)
              {
                 if (virtual2basic[j] == k) 
                 {
                    processesAttributes[k].speed += processesAttributes[j].speed;
                    probes++;
                 }
              }

              processesAttributes[k].speed /= probes;
              processorsAttributes[i] = processesAttributes[k];
              number_of_processes_per_node[i] = probes;
          }

          if (MPC_DEBUG > 1)
          {
             MPC_Debug_printf("host: Number of processes per node %d probes %d\n", number_of_processes_per_node[0], probes);
          }

          free(processesAttributes);

          if (MPC_DEBUG > 1)
          {
             MPC_Debug_printf("host: Building cluster stacks\n");
          }

          cluster_num = number_of_unique_nodes + 1;
          cluster_pointer = 0;
          cluster_process_num = 0;

          cluster_stack = (int(*)[2])calloc(
                                     cluster_num,
                                     sizeof(int[2])
          );

          if (cluster_stack == NULL)
          {
             return MPC_ERR_NOMEM;
          }

          cluster_stack[0][0] = 0;

          for (i = 0; i < number_of_unique_nodes; i++)
          {
              cluster_pointer++;
              cluster_stack[cluster_pointer][0] = cluster_process_num;
              cluster_process_num += number_of_processes_per_node[i];
              cluster_stack[cluster_pointer][1] = cluster_process_num - 1;
          }

          cluster_stack[0][1] = cluster_process_num - 1;

          free(virtual2basic);

          if (MPC_DEBUG > 1)
          {
             MPC_Debug_printf("host: cluster stacks built\n"); fflush(stdin);
          }
       }

       {
          int i, j, k, rep_count, source, dest, block_size;

          double start, end, speeds[MPC_TOPO_REPS][MPC_TRANSFER_DATA_RANGE];

          char *buff;

          MPI_Status status;

          if ((mpi_comm_of_nodes != MPI_COMM_NULL) 
              && (me != host
             )
          )
          {
             processorsAttributes = (MPC_tNodeAttribute_i*)calloc(
                                                           number_of_unique_nodes,
                                                           sizeof(MPC_tNodeAttribute_i)
             );

             if (processorsAttributes == NULL)
             {
                return MPC_ERR_NOMEM;
             }

             number_of_processes_per_node = (int*)calloc(
                                                  number_of_unique_nodes, 
                                                  sizeof(int)
             );

             if (number_of_processes_per_node == NULL)
             {
                return MPC_ERR_NOMEM;
             }
          }

          if (mpi_comm_of_nodes != MPI_COMM_NULL)
          {
             rc = MPI_Bcast(
                     processorsAttributes, 
                     number_of_unique_nodes, 
                     attrtype, 
                     host, 
                     mpi_comm_of_nodes
             );

             rc = I2C(rc);

             if (rc != MPC_OK)
             {
                return rc;
             }

             rc = MPI_Bcast(
                     number_of_processes_per_node, 
                     number_of_unique_nodes, 
                     MPI_INT, 
                     host, 
                     mpi_comm_of_nodes
             );

             rc = I2C(rc);

             if (rc != MPC_OK)
             {
                return rc;
             }
          }
  
          rc = MPI_Bcast(
                  &cluster_num, 
                  1, 
                  MPI_INT, 
                  host, 
                  mpi_comm_world
          );

          rc = I2C(rc);

          if (rc != MPC_OK)
          {
             return rc;
          }

          if (me != host)
          {
             cluster_stack = (int(*)[2])calloc(cluster_num, sizeof(int[2]));
             
             if (cluster_stack == NULL)
             {
                return MPC_ERR_NOMEM;
             }
          }

          rc = MPI_Bcast(
                  cluster_stack, 
                  cluster_num*2, 
                  MPI_INT, 
                  host, 
                  mpi_comm_world
          );
  
          rc = I2C(rc);

          if (rc != MPC_OK)
          {
             return rc;
          }

          if (MPC_DEBUG > 1)
          {
             MPC_Debug_printf("me=%d: Broadcast of cluster stacks, number of processes per node... complete\n", me);
          }

          source = cluster_stack[0][0];
          dest = cluster_stack[0][1];

          for (rep_count = 0; rep_count < MPC_TOPO_REPS; rep_count++) 
          {
             for (k=0, block_size = MPC_SIZE_OF_DATA_PACK;
                  k < MPC_TRANSFER_DATA_RANGE; 
                  k++, block_size *= MPC_SIZE_OF_DATA_PACK
             )
             {
                 buff = (char*)malloc(block_size*sizeof(char));

                 if (buff == NULL)
                 {
                    return MPC_ERR_NOMEM;
                 }

                 rc = MPI_Barrier(mpi_comm_world);

                 rc = I2C(rc);

                 if (rc != MPC_OK)
                 {
                    return rc;
                 }

                 if (source != dest) 
                 {
                    if (me == source)
                    {
                       start = MPI_Wtime();

                       rc = MPI_Send(
                               buff,
                               block_size,
                               MPI_CHAR,
                               dest,
                               host,
                               mpi_comm_world
                       );

                       rc = I2C(rc);

                       if (rc != MPC_OK)
                       {
                          return rc;
                       }

                       rc = MPI_Recv(
                               buff,
                               block_size,
                               MPI_CHAR,
                               dest,
                               host,
                               mpi_comm_world, 
                               &status
                       );

                       rc = I2C(rc);

                       if (rc != MPC_OK)
                       {
                          return rc;
                       }

                       end = MPI_Wtime();

                       speeds[rep_count][k] = 2.0*block_size/(end-start);
                    }

                    if (me == dest)
                    {
                       rc = MPI_Recv(
                               buff,
                               block_size,
                               MPI_CHAR,
                               source,
                               host,
                               mpi_comm_world,
                               &status
                       );

                       rc = I2C(rc);

                       if (rc != MPC_OK)
                       {
                          return rc;
                       }

                       rc = MPI_Send(
                               buff,
                               block_size,
                               MPI_CHAR,
                               source,
                               host,
                               mpi_comm_world
                       );

                       rc = I2C(rc);

                       if (rc != MPC_OK)
                       {
                          return rc;
                       }
                    }
                 }
                 else 
                 {
                     char *bufff = (char*)malloc(block_size*sizeof(char));

                     if (bufff == NULL)
                     {
                        return MPC_ERR_NOMEM;
                     }

                     start = MPI_Wtime();

                     memcpy(bufff, buff, block_size);

                     end = MPI_Wtime();

                     speeds[rep_count][k] = block_size/(end-start);

                     free(bufff);
                 }

                 free(buff);
             }

             rc = MPI_Bcast(
                     speeds[rep_count], 
                     MPC_TRANSFER_DATA_RANGE, 
                     MPI_DOUBLE,
                     source, 
                     mpi_comm_world
             );

             rc = I2C(rc);

             if (rc != MPC_OK)
             {
                return rc;
             }
          }

          if (MPC_DEBUG > 1)
          {
             MPC_Debug_printf("me=%d: Top level runtime platform parameters gathered\n", me);
          }

          if (mpi_comm_of_nodes != MPI_COMM_NULL) 
          {
             for (rep_count = 1; rep_count < MPC_TOPO_REPS; rep_count++)
             {
                for (k = 0; k < MPC_TRANSFER_DATA_RANGE; k++)
                {
                   speeds[0][k] += speeds[rep_count][k];
                }
             }

             for (k = 0; k < MPC_TRANSFER_DATA_RANGE; k++)
             {
                speeds[0][k] /= MPC_TOPO_REPS;
             }
          
             /*
             * If Iam the host
             */
             if (me == host)
             {
                /*
                * Fill in the topology information
                */
                int l;

                for (l = 0; l < MPC_TRANSFER_DATA_RANGE; l++) 
                {
                    link_nodes[lncount].linkspeed[l] = 1.0 
                                                       /
                                                       speeds[0][k]
                    ;
                }

                link_nodes[lncount].level = 0;

                link_nodes[lncount].top = NULL;

                if (save_to_topofile == 1)
                {
                   for (k = 0; 
                        k < MPC_TRANSFER_DATA_RANGE;
                        k++
                   )
                   {
                        fprintf(
                           wtopofp, 
                           "c%d%c", 
                           (int)(speeds[0][k]), 
                           (k != MPC_TRANSFER_DATA_RANGE - 1) ? ' ' : '\n'
                        );
                   }
                }

                /*
                * Fill the environment buffer
                */
                {
                   int is, ks = *penv_length;

                   *penv_length += 3+MPC_TRANSFER_DATA_RANGE;

                   *penvironment = realloc(
                                   (*penvironment),
                                   sizeof(Int)*(*penv_length)
                   );

                   if (*penvironment == NULL)
                   {
                      return MPC_ERR_NOMEM;
                   }

                   (*penvironment)[ks++] = MPC_LINK_NODE_TAG;

                   for (is = 0; is < MPC_TRANSFER_DATA_RANGE; is++)
                   {
                      (*penvironment)[ks++] = speeds[0][is];
                   }

                   (*penvironment)[ks++] = link_nodes[lncount].level;
                   (*penvironment)[ks++] = MPC_NO_PARENT_TAG;
                }

                lncount++;

                if (MPC_DEBUG > 1)
                {
                   MPC_Debug_printf("host: Top level (root, level 0) runtime platform parameters gathered and written to the file if necessary\n");
                }
             }
          }

          for (j = 1, i = 0; j < cluster_num; j++)
          {
            if (mpi_comm_of_nodes != MPI_COMM_NULL)
            {
                /*
                * If Iam the host
                */
                if (me == host)
                {
                   /*
                   * Fill in the topology information
                   */
                   int ni;
                   int processscalability = MPC_Nearest_i(
                                               number_of_processes_per_node[i]
                                               *
                                               processorsAttributes[i].scalability
                   );

                   groups[i].maxproc = number_of_processes_per_node[i];
                   groups[i].numproc = 0;
                   groups[i].scalability = processscalability;
                   groups[i].processor_power = (int)(processorsAttributes[i].speed)+1;
                   groups[i].max_nodepower = groups[i].processor_power/MPC_POWER_COEFF;
                   groups[i].nodeweights = (double*)malloc(
                                                    sizeof(double)
                                                    *
                                                    groups[i].maxproc
                   );

                   if (groups[i].nodeweights == NULL)
                   {
                      return MPC_ERR_NOMEM;
                   }

                   groups[i].nodenumbers = (int*)malloc(
                                                 sizeof(int)
                                                 *
                                                 groups[i].maxproc
                   );

                   if (groups[i].nodenumbers == NULL)
                   {
                      return MPC_ERR_NOMEM;
                   }

                   for (ni = 0; ni < groups[i].maxproc; ni++)
                   {
                       groups[i].nodenumbers[ni] = total_processes++;
                       groups[i].nodeweights[ni] = 0.0;
                   }

                   /*
                   * Fill the environment buffer
                   */
                   {
                      int is, ks = *penv_length;

                      (*penvironment)[0]++;

                      *penv_length += 3+groups[i].maxproc;

                      *penvironment = realloc(
                                      (*penvironment),
                                      sizeof(Int)*(*penv_length)
                      );

                      if (*penvironment == NULL)
                      {
                         return MPC_ERR_NOMEM;
                      }

                      (*penvironment)[ks++] = groups[i].maxproc;
                      (*penvironment)[ks++] = groups[i].scalability;
                      (*penvironment)[ks++] = groups[i].processor_power;

                      for (is = 0; is < groups[i].maxproc; is++)
                      {
                          (*penvironment)[ks++] = groups[i].nodenumbers[is];
                      }
                   }

                   if (save_to_topofile == 1)
                   {
                      fprintf(
                         wtopofp, 
                         "#%s\ns%d p%d n%d ", 
                          nodenames[basic_ranks[i]],
                          processscalability,
                          (int)(processorsAttributes[i].speed)+1,
                          number_of_processes_per_node[i]
                      );
                   }

                   if (MPC_DEBUG > 1)
                   {
                      MPC_Debug_printf("host: level %d runtime platform parameters gathered and written to the file if necessary\n", i+1);
                   }
                }
            }

            source = cluster_stack[j][0];
            dest = cluster_stack[j][1];

            for (rep_count = 0; rep_count < MPC_TOPO_REPS; rep_count++) 
            {
                for (k = 0, block_size = MPC_SIZE_OF_DATA_PACK;
                     k < MPC_TRANSFER_DATA_RANGE;
                     k++, block_size *= MPC_SIZE_OF_DATA_PACK
                )
                {
                     buff = (char*)malloc(block_size*sizeof(char));

                     if (buff == NULL)
                     {
                        return MPC_ERR_NOMEM;
                     }

                     rc = MPI_Barrier(mpi_comm_world);

                     rc = I2C(rc);
    
                     if (rc != MPC_OK)
                     {
                        return rc;
                     }

                     if (source != dest)
                     {
                        if (me == source)
                        {
                           start = MPI_Wtime();

                           rc = MPI_Send(
                                   buff,
                                   block_size,
                                   MPI_CHAR,
                                   dest,
                                   host,
                                   mpi_comm_world
                           );

                           rc = I2C(rc);

                           if (rc != MPC_OK)
                           {
                              return rc;
                           }

                           rc = MPI_Recv(
                                   buff,
                                   block_size,
                                   MPI_CHAR,
                                   dest,
                                   host,
                                   mpi_comm_world,
                                   &status
                           );

                           rc = I2C(rc);

                           if (rc != MPC_OK)
                           {
                              return rc;
                           }

                           end = MPI_Wtime();

                           speeds[rep_count][k] = 2.0*block_size/(end-start);
                        }

                        if (me == dest)
                        {
                            rc = MPI_Recv(
                                    buff,
                                    block_size,
                                    MPI_CHAR,
                                    source,
                                    host,
                                    mpi_comm_world,
                                    &status
                            );

                            rc = I2C(rc);

                            if (rc != MPC_OK)
                            {
                               return rc;
                            }

                            rc = MPI_Send(
                                    buff,
                                    block_size,
                                    MPI_CHAR,
                                    source,
                                    host,
                                    mpi_comm_world
                            );

                            rc = I2C(rc);

                            if (rc != MPC_OK)
                            {
                               return rc;
                            }
                        }
                     }
                     else 
                     {
                         char *bufff = (char*)malloc(block_size*sizeof(char));

                         if (bufff == NULL)
                         {
                            return MPC_ERR_NOMEM;
                         }

                         start = MPI_Wtime();

                         memcpy(bufff, buff, block_size);

                         end = MPI_Wtime();

                         speeds[rep_count][k] = block_size/(end-start);

                         free(bufff);
                     }

                     free(buff);
                }

                rc = MPI_Bcast(
                        speeds[rep_count], 
                        MPC_TRANSFER_DATA_RANGE, 
                        MPI_DOUBLE,
                        source, 
                        mpi_comm_world
                );

                rc = I2C(rc);

                if (rc != MPC_OK)
                {
                   return rc;
                }
            }

            if (MPC_DEBUG > 1)
            {
               MPC_Debug_printf("me=%d: level %d runtime platform parameters gathered\n", me, i+1);
            }

            if (mpi_comm_of_nodes != MPI_COMM_NULL) 
            {
               for (rep_count = 1; rep_count < MPC_TOPO_REPS; rep_count++)
               {
                   for (k = 0; k < MPC_TRANSFER_DATA_RANGE; k++)
                   {
                       speeds[0][k] += speeds[rep_count][k];
                   }
               }

               for (k = 0; k < MPC_TRANSFER_DATA_RANGE; k++)
               {
                   speeds[0][k] /= MPC_TOPO_REPS;
               }
              
               /*
               * If Iam the host
               */
               if (me == host)
               {
                  for (k = 0; k < MPC_TRANSFER_DATA_RANGE; k++)
                  {
                      link_nodes[lncount].linkspeed[k] = 1.0
                                                         /
                                                         speeds[0][k]
                      ;
                  }

                  link_nodes[lncount].level = 1;

                  link_nodes[lncount].top = link_nodes;

                  groups[i].link_descr = link_nodes+lncount;

                  /*
                  * Fill the environment buffer
                  */
                  {
                     int is, ks = *penv_length;                     
  
                     *penv_length += 2+MPC_TRANSFER_DATA_RANGE;

                     *penvironment = realloc(
                                     (*penvironment),
                                     sizeof(Int)*(*penv_length)
                     );

                     if (*penvironment == NULL)
                     {
                        return MPC_ERR_NOMEM;
                     }
  
                     for (is = 0; is < MPC_TRANSFER_DATA_RANGE; is++)
                     {
                        (*penvironment)[ks++] = speeds[0][is];
                     }

                     (*penvironment)[ks++] = link_nodes[lncount].level;
                     (*penvironment)[ks++] = 1;
                  }

                  lncount++;

                  if (save_to_topofile == 1)
                  {
                     for (k = 0;
                          k < MPC_TRANSFER_DATA_RANGE;
                          k++
                     )
                     {
                       fprintf(
                          wtopofp, 
                          "c%d%c", 
                          (int)(speeds[0][k]),
                          ((k != MPC_TRANSFER_DATA_RANGE - 1) ? ' ' : '\n')
                       );
                     }
                  }

                  if (MPC_DEBUG > 1)
                  {
                     MPC_Debug_printf("host: level %d runtime platform parameters gathered and written to the file if necessary\n", i+1);
                  }
               }
            }

            i++;
        }

        if (mpi_comm_of_nodes != MPI_COMM_NULL) 
        {
           free(number_of_processes_per_node);
           free(processorsAttributes);
           free(nodenames);

           /*
           * If Iam the host
           */
           if (me == host)
           {
              if (save_to_topofile == 1)
              {
                 fclose(wtopofp);
                 wtopofp = NULL;
              }
           }
        }

      }
    }

    if (MPC_DEBUG > 1)
    {
       MPC_Debug_printf("me=%d: runtime platform parameters generation complete\n", me);
    }

    rc = MPI_Type_free(&attrtype);

    rc = I2C(rc);

    if (rc != MPC_OK)
    {
       return rc;
    }

    rc = MPI_Group_free(
            &mpi_group_world
    );

    rc = I2C(rc);

    if (rc != MPC_OK)
    {
       return rc;
    }

    rc = MPI_Group_free(
            &mpi_group_of_nodes
    );

    rc = I2C(rc);

    if (rc != MPC_OK)
    {
       return rc;
    }

    if (mpi_comm_of_nodes != MPI_COMM_NULL)
    {
       rc = MPI_Comm_free(
               &mpi_comm_of_nodes
       );

       rc = I2C(rc);

       if (rc != MPC_OK)
       {
          return rc;
       }
    }

    free(basic_ranks);
    free(cluster_stack);

    /*
    * Local environment is built
    */
    if (me == host)
    {
       int i;

       MPC_Local_env_built = 1;
 
       free(link_nodes);

       for (i = 0; i < ngroups; i++)
       {
          free(groups[i].nodeweights);
          free(groups[i].nodenumbers);
       }

       free(groups);
       
       MPC_Processes_num = total_processes;

       if (MPC_DEBUG > 1)
       {
          MPC_Debug_printf("host: total number of processes %d, comm world size %d (both must be equal)\n", MPC_Processes_num, commworldsize);
       }
    }

    return MPC_OK;
}

/************************ MPC_Initialize_topology_environment ************************************/

