
   /************************************************/
   /* Implementation of Partitioning Interfaces of */
   /* Sets using processor graphs with speed and   */
   /* memory                                       */
   /*                                              */
   /* Revision history                             */
   /* 01-07-2003  --      Initial version          */
   /************************************************/

   #include <stdio.h>
   #include <stdlib.h>
   #include <string.h>

   #include <math.h>
   #include <hmpi.h>

   static int _HMPI_Bisection_count = 0;

   /*-----------------------------------------------------*/

   int 
   __HMPI_Distribute_with_single_number_for_speed
   (
      int n,
      int p,
      const double *s,
      double *npd
   )
   {
       int i, left, rc, sum = 0;
       int* npp = (int*)malloc(
                        sizeof(int)
                        *
                        p
       );

       if (npp == NULL)
       {
          return MPC_ERR_NOMEM;
       }

       for (i = 0; i < p; i++)
       {
	   sum += npd[i];
       }

       left = n - sum;

       if (left > 0)
       {
          rc = __HMPI_Number_of_elements_proportional_to_speed(
                     p,
		     left,
		     s,
		     npp
          );

	  if (rc != HMPI_OK)
          {
             return rc;
          }

          for (i = 0; i < p; i++)
          {
	      npd[i] += npp[i];
          }
       }
       else
       {
          rc = __HMPI_Number_of_elements_proportional_to_speed(
                     p,
		     sum - n,
		     s,
		     npp
          );

	  if (rc != HMPI_OK)
          {
             return rc;
          }

          for (i = 0; i < p; i++)
          {
	      npd[i] -= npp[i];
          }
       }

       free(npp);

       return HMPI_OK;
   }

   /*-----------------------------------------------------*/

   int __HMPI_Recursive_bisection_middle_region
   (
       int p,
       int pn,
       const double *speeds,
       const int *psizes,
       int n,
       double slopei,
       double slopef,
       double *speeds_opt,
       double *npd
   )
   {
       int i, j, rc;
       double sumd;
       double slope, slopeii, slopeff;
       double slope_1_plus_2;
       double *s = (double*)malloc(
                            sizeof(double)
                            *
                            p
       );

       if (s == NULL)
       {
          return MPC_ERR_NOMEM;
       }

       /*
       * Initialize npd
       */
       for (i = 0; i < p; i++)
       {
           npd[i] = 0.;
       }

       /*
	* Now use the straight line with half slope
        */
       slope_1_plus_2 = (double)(slopei+slopef)
                        /
                       (double)(1 - slopei*slopef);

       if (slope_1_plus_2 > 0)
       {
          slope = (double)(sqrt(1+pow(slope_1_plus_2, 2)) - 1)
                  /
                  (double)slope_1_plus_2;
       }

       if (slope_1_plus_2 < 0)
       {
          slope = (double)(sqrt(1+pow(slope_1_plus_2, 2)) + 1)
                  / 
                  (double)(-slope_1_plus_2);
       }

       if (HMPI_Debug_flag)
       {
          printf(
            "HMPI===> __HMPI_Recursive_bisection_middle_region: Slope is %0.10f, %0.10f, %0.10f\n", 
            slopei, slopef, slope
          );
       }

       for (i = 0; i < p; i++)
       {
           double slopep;
           double interceptp;
           double x;

           int intersection_point_found = 0;

           for (j = 0; j < (pn - 1); j++)
           {
              /*
              * Ignore problem sizes of 0 for the moment
              */
              if (psizes[i*pn + j] == 0)
              {
                 npd[i] = 0;
                 speeds_opt[i] = DBL_MAX;
                 s[i] = DBL_MAX;
                 continue;
              }

              /* 
              * This is a extra/bad experimental point
              *
              * If this is the last point, assume a constant function
              * for the jump
              */
              if (psizes[i*pn + j + 1] == psizes[i*pn + j])
              {
                 if (j == (pn - 2))
                 {
                    x = speeds[i*pn + j]
                        /
                        slope;

                    if (x >= 0)
                    {
                       npd[i] = x;
                       speeds_opt[i] = s[i] = speeds[i*pn + j];
                    }

                    break;
                 }

                 continue;
              }

              slopep = (speeds[i*pn + j + 1] - speeds[i*pn + j])
                       /
                       (double)(psizes[i*pn + j + 1] - psizes[i*pn + j]);

              interceptp = speeds[i*pn + j] - slopep*psizes[i*pn + j];

              x = (double)interceptp
                  /
                  (double)(slope - slopep);

              /*
              * The x-coordinate of the intersection is negative
              * Proceed to the next straight line in the 
              * functional model
              */
              if (x < 0)
              {
                 continue;
              }

              /*
              * The intersection point lies beyond the end points
              * of the straight line, so should we proceed to the next 
              * iteration?
              */
              if ((x < psizes[i*pn + j])
                  || (x > psizes[i*pn + j + 1]
                  )
              )
              {
                 continue;
              }
 
              intersection_point_found = 1;
              npd[i] = x;
              speeds_opt[i] = s[i] = x*slope;
              break;
           }

           if (intersection_point_found == 1)
           {
              continue;
           }

           /*
           * Start with a function with the same slope as the first line segment
           */
           slopep = (speeds[i*pn + 1] - speeds[i*pn])
                    /
                    (double)(psizes[i*pn + 1] - psizes[i*pn]);

           interceptp = speeds[i*pn] - slopep*psizes[i*pn];

           x = (double)interceptp
               /
               (double)(slope - slopep);

           if ((x >= 0)
               && (x <= psizes[i*pn]
              )
           )
           {
              npd[i] = x;
              s[i] = x*slope;
              speeds_opt[i] = s[i];
              continue;
           }

           /*
           *  Assume a constant function in the beginning
           *  for a problem size of 0 and the first experimental
           *  point
           */
           if (psizes[i*pn] != 0)
           {
              x = (double)speeds[i*pn]
                  /
                  (double)slope;

              if ((x >= 0)
                  && (x <= psizes[i*pn]
                 )
              )
              {
                 npd[i] = x;
                 speeds_opt[i] = s[i] = speeds[i*pn];
                 continue;
              }
           }

           /*
           * Try the function with the same slope as before in the end
           */
           slopep = (speeds[i*pn + pn - 1] - speeds[i*pn + pn - 2])
                    /
                    (double)(psizes[i*pn + pn - 1] - psizes[i*pn + pn - 2]);
           interceptp = speeds[i*pn + pn - 2] - slopep*psizes[i*pn + pn - 2];
           x = (double)interceptp
               /
               (double)(slope - slopep);

           if (x >= 0)
           {
              npd[i] = x;
              speeds_opt[i] = s[i] = x*slope;
              continue;
           }

           /*
           * Now assume a constant function in the end
           */
           x = (double)speeds[i*pn + pn - 1]
               /
               (double)slope;

           if (x >= 0)
           {
              npd[i] = x;
              speeds_opt[i] = s[i] = speeds[i*pn + pn - 1];
              continue;
           }

           if (x < 0)
           {
              printf("HMPI===> __HMPI_Recursive_bisection_middle_region: Panic, no intersection\n");
           }
       }

       /*
	* If the sum is equal to n, we have a 
	* perfect fit.
	*/
       sumd = 0.0;
       for (i = 0; i < p; i++)
       {
	   sumd += npd[i];
       }

       if (((floor(sumd)) == n)
           || ((ceil(sumd)) == n
          )
       )
       {
          int sum = 0;

          for (i = 0; i < p; i++)
          {
	     sum += floor(npd[i]);
          }

          if (sum == n)
          {
             return HMPI_OK;
          }

          for (i = 0; i < p; i++)
          {
             if (npd[i] <= 1)
             {
                continue;
             }

             npd[i] = npd[i] + 1;

             sum = 0;

             for (j = 0; j < p; j++)
             {
                sum += floor(npd[j]);
             }

             if (sum == n)
             {
                break;
             }
          }

          free(s);

          return HMPI_OK;
       }

       if (HMPI_Debug_flag)
       {
          printf("HMPI===> __HMPI_Recursive_bisection_middle_region: Sum is %0.6f\n", sumd);
       }

       /*
        * After HMPI_MAX_BISECTION_STEPS steps, we have not
        * arrived at a solution. It is known that for bisection, no more 
        * than 53 iterations are needed to obtain full single precision.
        * Distribute the rest of the elements of the set 
        * using the speeds at the current point.
        */
       if (_HMPI_Bisection_count == HMPI_MAX_BISECTION_STEPS)
       {
          for (i = 0; i < p; i++)
          {
             speeds_opt[i] = s[i];
          }

          rc = __HMPI_Distribute_with_single_number_for_speed(
                     n, 
                     p, 
                     s,
                     npd
          );

          if (rc != HMPI_OK)
          {
             return rc;
          }

          free(s);

          return HMPI_OK;
       }

       free(s);

       _HMPI_Bisection_count++;

       if (sumd > n)
       {
          slopeii = slopei;
	  slopeff = slope;

	  return __HMPI_Recursive_bisection_middle_region(
                     p,
		     pn,
		     speeds,
		     psizes,
                     n,
                     slopeii,
		     slopeff,
                     speeds_opt,
		     npd
          );
       }
       
       slopeii = slope;
       slopeff = slopef;

       return __HMPI_Recursive_bisection_middle_region(
                  p,
		  pn,
		  speeds,
		  psizes,
                  n,
                  slopeii,
		  slopeff,
                  speeds_opt,
		  npd
       );
   }
   
   /*-----------------------------------------------------*/

   int __HMPI_Speed_function_of_problem_size_with_mlimits
   (
       int p,
       int pn,
       const double *speeds,
       const int *psizes,
       const int *bounds,
       int n,
       double *speeds_opt,
       int *np
   )
   {
       int i, j, rc;
       int bound_exceeded = 0;
       double sumd = 0;

       _HMPI_Bisection_count  = 0;

       for (i = 0; i < p; i++)
       {
          sumd += bounds[i];
       }

       if (sumd < n)
       {	        
          printf(
            "Problem size %d cannot be solved, memory bounds "
            "on the number of elements that can be stored by "
            "each processor exceeded\n", n);
          return HMPI_ERR_PARTITION_SET;
       }

       if (sumd == n)
       {	        
          for (i = 0; i < p; i++)
          {
             np[i] = bounds[i];
          }

	  return HMPI_OK;
       }

       rc = __HMPI_Speed_function_of_problem_size(
                p,
                pn,
                speeds,
                psizes,
                n,
                speeds_opt,
                np
       );

       if (rc != HMPI_OK)
       {
          return rc;
       }

       for (i = 0; i < p; i++)
       {
           if (np[i] > bounds[i])
           {
              np[i] = bounds[i];

              bound_exceeded = 1;

              break;
           }
       }

       if (bound_exceeded == 1)
       {
          int k;
          int ind = 0;
          int sind = 0;
          int mind = 0;
          int *psizesm;
          double *speedsm;
          double *speedsm_opt;
          int *npm;
          int *boundsm;
          int nm = n - bounds[i];

          speedsm = (double*)malloc(
                             sizeof(double)
                             *
                             (p-1)
                             *
                             pn
          );        

          if (speedsm == NULL)
          {
             return MPC_ERR_NOMEM;
          }

          speedsm_opt = (double*)malloc(
                             sizeof(double)
                             *
                             (p-1)
                             *
                             pn
          );        

          if (speedsm_opt == NULL)
          {
             return MPC_ERR_NOMEM;
          }

          psizesm  = (int*)malloc(
                           sizeof(int)
                           *
                           (p-1)
                           *
                           pn
          );        

          if (psizesm  == NULL)
          {
             return MPC_ERR_NOMEM;
          }

          npm = (int*)malloc(
                      sizeof(int)
                      *
                      (p-1)
          );        

          if (npm == NULL)
          {
             return MPC_ERR_NOMEM;
          }

          boundsm = (int*)malloc(
                          sizeof(int)
                          *
                          (p-1)
          );        

          if (boundsm == NULL)
          {
             return MPC_ERR_NOMEM;
          }

          for (j = 0; j < p; j++)
          {
             if (j == i)
             {
                continue;
             }

             npm[ind] = np[j];
             boundsm[ind] = bounds[j];
             ind++;

             for (k = 0; k < pn; k++)
             {
                 speedsm[sind++] = speeds[j*pn + k];
                 psizesm[mind++] = psizes[j*pn + k];
             }
          }

          rc = __HMPI_Speed_function_of_problem_size_with_mlimits(
                   p-1,
                   pn,
                   speedsm,
                   psizesm,
                   boundsm,
                   nm,
                   speedsm_opt,
                   npm
          );

          if (rc != HMPI_OK)
          {
             return rc;
          }

          for (ind  = 0, j = 0; j < p; j++)
          {
             if (j == i)
             {
                continue;
             }

             np[j] = npm[ind];
             speeds_opt[j] = speedsm_opt[ind];
             ind++;
          }

          free(speedsm);
          free(speedsm_opt);
          free(psizesm);
          free(boundsm);
          free(npm);
       }

       return HMPI_OK;
   }

   /*-----------------------------------------------------*/

   int __HMPI_Speed_function_of_problem_size
   (
       int p,
       int pn,
       const double *speeds,
       const int *psizes,
       int n,
       double *speeds_opt,
       int *np
   )
   {
       int istart = p;
       int negative_x_intersection = 0;
       int rc, ind, i, j;
       double temp, sumd;
       double slopei, slopef;
       double *npd = (double*)malloc(
                              sizeof(double)
                              *
                              p
       );

       if (npd == NULL)
       {
          return MPC_ERR_NOMEM;
       }

       _HMPI_Bisection_count = 0;

       /*
        * Initialize npd
        */
       for (i = 0; i < p; i++)
       {
           npd[i] = 0.;
       }

       /*
	* The functions may start from the problem size of 0 
	*/
       for (i = 0; i < p; i++)
       {
          if (psizes[i*pn] != 0)
          {
             istart = i;
             break;
          }
       }

       if (istart != p)
       {
          temp = (double)speeds[istart*pn]
	         /
	         (double)psizes[istart*pn];
          ind = istart;
          for (i = istart+1; i < p; i++)
          {
              double temps;

              /*
              * Ignore the point where the problem size is 0
              */
              if (psizes[i*pn] == 0)
              {
                 continue;
              }

              temps = (double)speeds[i*pn]
	              /
	              (double)psizes[i*pn];

              if (temp < temps)
              {
                 temp = temps;
                 ind = i;
              }
          }
       }
       else
       {
          ind = 0;
       }

       /*
	* Solve the equations 
	* y = (maximum slope)*x and 
	* y0 = c0, y1 = c1, ... for points
	* x0, x1, x2, ...,xp-1
	*/
       npd[ind] = psizes[ind*pn];
       speeds_opt[ind] = speeds[ind*pn];
       for (i = 0; i < p; i++)
       {
           double xj;

	   if (i == ind)
           {
              continue;
           }

           /*
           * Ignore problem sizes of 0 for the moment
           */
           if (psizes[i*pn] == 0)
           {
              npd[i] = 0;
              speeds_opt[i] = speeds[i*pn];
              break;
           }

           xj = (
	          (double)psizes[ind*pn]
	          /
	          (double)speeds[ind*pn]
                )
	        *
	        speeds[i*pn];

           npd[i] = xj;
           speeds_opt[i] = speeds[i*pn];
       }

       /*
	* If the sum is equal to n, we have a 
	* perfect fit.
	*/
       sumd = 0.0;
       for (i = 0; i < p; i++)
       {
	  sumd += npd[i];
       }

       if (((floor(sumd)) == n)
           || ((ceil(sumd)) == n
           )
       )
       {
          int sum = 0;

          for (i = 0; i < p; i++)
          {
	     sum += floor(npd[i]);
          }

          if (sum == n)
          {
             for (i = 0; i < p; i++)
             {
                np[i] = floor(npd[i]);
             }

             return HMPI_OK;
          }

          for (i = 0; i < p; i++)
          {
             if (npd[i] <= 1)
             {
                continue;
             }

             npd[i] = npd[i] + 1;

             sum = 0;

             for (j = 0; j < p; j++)
             {
                 sum += floor(npd[j]);
             }

             if (sum == n)
             {
                break;
             }
          }

          for (i = 0; i < p; i++)
          {
              np[i] = floor(npd[i]);
          }

          free(npd);

          return HMPI_OK;
       }

       /*
       * The intersection points lie before the starting 
       * experimental point. So use the single number speeds
       * This is the only thing we can do at this point.
       */
       if (sumd > n)
       {
          double *s = (double*)malloc(
			       sizeof(double)
			       *
			       p
          );

	  if (s == NULL)
          {
             return MPC_ERR_NOMEM;
          }

	  for (i = 0; i < p; i++)
          {
              s[i] = speeds[i*pn];
          }

          rc = __HMPI_Number_of_elements_proportional_to_speed(
                      p,
		      n,
		      s,
		      np
          );

	  if (rc != HMPI_OK)
          {
             return rc;
          }

          free(npd);
	  free(s);

          return HMPI_OK;
       }
          
       if (istart != p)
       {
          slopei = (double)speeds[ind*pn]
                   /
	           (double)psizes[ind*pn];
       }
       else
       {
          slopei = DBL_MAX;
       }

       /*
	* Start with a straight line passing
	* with minimum slope at last point
        *
        * Actually this is ridiculous. The 
        * last point can't have the problem size 0.
        * But we take care of this situation as best as we can.
	*/
       istart = p;
       for (i = 0; i < p; i++)
       {
          if (psizes[i*pn + pn - 1] != 0)
          {
             istart = i;
             break;
          }
       }

       if (istart != p)
       {
          temp = (double)speeds[istart*pn + pn - 1]
                 /
                 (double)psizes[istart*pn + pn - 1];
          ind = istart;
          for (i = istart+1; i < p; i++)
          {
              double temps = (double)speeds[i*pn + pn - 1]
		             /
		             (double)psizes[i*pn + pn - 1];
              if (temp > temps)
              {
                 temp = temps;
                 ind = i;
              }
          }
       }
       else
       {
          ind =  0;
       }

       /*
	* Solve the equations for the last point
	* y = (minimum slope)*x and for the processors
	* y0 = b0*x+c0, y1 = b1*x1+c1, ... for points
	* x0, x1, x2, ...,xp-1
        *
        * Initialize npd
        */
       for (i = 0; i < p; i++)
       {
           npd[i] = 0.;
       }

       npd[ind] = psizes[ind*pn + pn - 1];
       speeds_opt[ind] = speeds[ind*pn + pn - 1];
       for (i = 0; i < p; i++)
       {
           if (i == ind)
           {
              continue;
           }

	   {
              double slopep, slopeo;
              double interceptp;
              double x, y;

              /*
              * Ignore problem sizes of 0 for the moment
              */
              if (psizes[i*pn + pn - 2] == 0)
              {
                 npd[i] = 0;
                 speeds_opt[i] = speeds[i*pn + pn - 2];
                 continue;
              }

              /*
              * This is an extra or bad experimental point
              */
              if (psizes[i*pn + pn - 1] == psizes[i*pn + pn - 2])
              {
                 /*
                 * Replace this jump by constant function
                 */
                 slopeo = speeds_opt[ind]
                          /
                          (double)npd[ind];

                 x = speeds[i*pn + pn - 2]
                     /
                     slopeo;

                 if (x >= 0)
                 {
                    npd[i] = x;
                    speeds_opt[i] = speeds[i*pn + pn - 2];
                 }

                 if (x < 0)
                 {
                    break;
                 }

                 continue;
              }

              slopep = (speeds[i*pn + pn - 1] - speeds[i*pn + pn - 2])
                       /
                       (double)(psizes[i*pn + pn - 1] - psizes[i*pn + pn - 2]);

              interceptp = speeds[i*pn + pn - 2] - slopep*psizes[i*pn + pn - 2];

              slopeo = speeds_opt[ind]
                       /
                       (double)npd[ind];

	      x = (double)interceptp
	          /
	          (double)(slopeo - slopep);

              /*
              * The x-coordinate of the intersection is negative
              */
              if (x < 0)
              {
                 negative_x_intersection = 1;
                 break;
              }

	      npd[i] = x;
              speeds_opt[i] = x*slopeo;
           }
       }

       if (negative_x_intersection == 0)
       {
          /*
	   * If the sum is equal to n, we have a 
	   * perfect fit.
	   */
          sumd = 0.0;
          for (i = 0; i < p; i++)
          {
	      sumd += npd[i];
          }

          if (((floor(sumd)) == n)
              || ((ceil(sumd)) == n
              )
          )
          {
             int sum = 0;

             for (i = 0; i < p; i++)
             {
	        sum += floor(npd[i]);
             }

             if (sum == n)
             {
                for (i = 0; i < p; i++)
                {
                   np[i] = floor(npd[i]);
                }
   
                return HMPI_OK;
             }
   
             for (i = 0; i < p; i++)
             {
                if (npd[i] <= 1)
                {
                   continue;
                }

                npd[i] = npd[i] + 1;

                sum = 0;

                for (j = 0; j < p; j++)
                {
                   sum += floor(npd[j]);
                }
   
                if (sum == n)
                {
                   break;
                }
             }
   
             for (i = 0; i < p; i++)
             {
                   np[i] = floor(npd[i]);
             }
   
             free(npd);
   
             return HMPI_OK;
          }

          /*
          * The functions constructed are inadequate. More
          * experimental points are required. 
          * Should we alert the user?
          */
          if (sumd < n)
          {
             double *s = (double*)malloc(
		      	          sizeof(double)
			          *
			          p
             );

	     if (s == NULL)
             {
                return MPC_ERR_NOMEM;
             }

	     for (i = 0; i < p; i++)
             {
                 s[i] = (
	                  (double)speeds[ind*pn + pn - 1]
		          /
		          (double)npd[ind]
		        ) 
		        *
		        npd[i];
                 speeds_opt[i] = s[i];
             }
   
             rc = __HMPI_Number_of_elements_proportional_to_speed(
                         p,
	   	         n,
		         s,
	   	         np
             );

	     if (rc != HMPI_OK)
             {
                return rc;
             }

	     free(s);
	     free(npd);

             return HMPI_OK;
          }
       }

       slopef = (double)speeds[ind*pn + pn - 1]
                /
	        (double)psizes[ind*pn + pn - 1];

       /*
	* Use recursive Bisection to get a perfect fit
	*/
       rc = __HMPI_Recursive_bisection_middle_region(
            p,
            pn,
            speeds,
            psizes,
	    n,
            slopei, 
	    slopef,
            speeds_opt,
            npd
       );

       if (rc != HMPI_OK)
       {
          return rc;
       }

       for (i = 0; i < p; i++)
       {
          np[i] = floor(npd[i]);
       }

       free(npd);

       return HMPI_OK;
   }

   /*-----------------------------------------------------*/

