
    #include <math.h>
    #include <stdio.h>
    #include <string.h>
    #include <stdlib.h>
    #include <sys/time.h>

    #include "mxm_i.h"

    /*----------------------------------------------------------------------------*/

    int InitializeMatricesLocal
    (
        int m,
        int n,
        double *a,
        double *b,
        double *c
    )
    {
        int i;

        for (i = 0; i < (m*n); i++)
        {
            a[i] = MXM_CONSTANT_NUMBER1;
            b[i] = MXM_CONSTANT_NUMBER2;
            c[i] = 0.0;
        }

        return HMPI_OK;
    }

    /*-----------------------------------------------------------------------------------------*/

    int dgemm_local_optimized(
        const int *myh,
        const int *myw,
        const int *kb,
        double    *A,
        double    *B,
        double    *C,
        double *etime
    )
    {
        int m, n, k, rc, lda, ldb, ldc;

        char TRANSAC   = 'N';
        char *TRANSA   = &TRANSAC;
        char TRANSBC   = 'N';
        char *TRANSB   = &TRANSBC;

        double tstart, tend;
        double  ALPHA  = 1.;
        double  BETA   = 1.;

	*etime = 0.;

        if ((*myw == 0)
            || (*myh == 0
           )
        )
        {
           return 0;
        }

        m = *myh;
        n = *myw;
        k = *kb;

        /*
        * C := alpha*op( A )*op( B ) + beta*C
        */
        tstart = MPI_Wtime();

        lda = ((m > 1) ? m : 1);
        ldb = ((k > 1) ? k : 1);
        ldc = ((m > 1) ? m : 1);

        if (CHECK_ARGUMENTS > 0)
        {
           int i;
           fprintf(dofp, "A[]:\n");
           for (i = 0; i < (m*k); i++)
           {
              fprintf(dofp, "%3f ", A[i]);
           }
           fprintf(dofp, "\n");
           fflush(dofp);
        }

        if (CHECK_ARGUMENTS > 0)
        {
           int i;
           fprintf(dofp, "B[]:\n");
           for (i = 0; i < (k*n); i++)
           {
              fprintf(dofp, "%3f ", B[i]);
           }
           fprintf(dofp, "\n");
           fflush(dofp);
        }

        dgemm_(
              TRANSA,
              TRANSB,
              &m,
              &n,
              &k,
              &ALPHA,
              A,
              &lda,
              B,
              &ldb,
              &BETA,
              C,
              &ldc
        );

        tend = MPI_Wtime();

        *etime = tend - tstart;

        return 0;
    }

    /*-----------------------------------------------------------------------------------------*/

    int dgemm_local_noptimized(
        const int *myh,
        const int *myw,
        const int *kb,
        double    *A,
        double    *B,
        double    *C,
        double *etime
    )
    {
        int    i, j, m, n, k, kk, rc;

        double tstart, tend;

	*etime = 0.;

        if ((*myw == 0)
            || (*myh == 0
           )
        )
        {
           return 0;
        }

        m = *myh;
        n = *myw;
        k = *kb;

        /*
        * C := alpha*op( A )*op( B ) + beta*C
        */
        tstart = MPI_Wtime();

        for (kk = 0; kk < k; kk++)
        {
            for (i = 0; i < m; i++)
            {
                double val = A[i*k + kk];
                for (j = 0; j < n; j++)
                {
                    C[i*n+j] += val
                                *
                                B[kk*n + j];
                }
            }
        }

        tend = MPI_Wtime();

        *etime = tend - tstart;

        return 0;
    }

    /*-----------------------------------------------------------------------------------------*/

    /*
    * At the moment, only dense square matrices are implemented
    *
    * And COLUMN BASED distribution is assumed
    */
    int Execute_scalapack_outer_product_algo(
        const int me,
        const int myi,
        const int myj,
        const int n,
        const int p,
        const int q,
        const int *w,
        const int *h,
        MPI_Comm  mxmcomm
    )
    {
        int    i, j, k, rc;

        double *A, *B, *C, etime, start, end;

        /*
        * Some constants used in the main algorithm loop
        */
        int NBYTES = sizeof(double)*kb;
        int MATELEMSIZE = kb*kb;

        /*
        * My local store is a matrix of size lda x ldb
        */
        int    myh = h[HMPI_RECT_INDEX(myi, myj, myi, myj, p, q)];
        int    myw = w[HMPI_RECT_INDEX(myi, myj, myi, myj, p, q)];
        int    lda = myh*kb;
        int    ldb = myw*kb;

        /*
        * Temporary buffers for A and B
        */
        double *WA   = NULL;
        double *WB   = NULL;

        /*
        * Unfortunately we have to send vertical slices of matrix A;
        * We need to prepare the vertical slice.
        * These temporary copies are required as the elements are
        * stored in row-major order
        */
        WA = (double*)malloc(
                      sizeof(double)
                      *
                      (lda*kb)
        );

        if (WA == NULL)
        {
           printf("me=%d: No memory to allocate the pivot column, WA\n", me);
        }

        WB = (double*)malloc(
                      sizeof(double)
                      *
                      (ldb*kb)
        );

        if (WB == NULL)
        {
           printf("me=%d: No memory to allocate the pivot row, WB\n", me);
        }

        /*
        * Allocate the matrices A, B, C
        */
        A = (double*)malloc(
                     sizeof(double)
                     *
                     (lda*ldb)
        );

        if (A == NULL)
        {
           fprintf(dofp, "(%d, %d):Cannot allocate A, m=%d, k=%d\n", myi, myj, lda, ldb);
           fflush(dofp);
           return -1;
        }

        B = (double*)malloc(
                     sizeof(double)
                     *
                     (lda*ldb)
        );

        if (B == NULL)
        {
           fprintf(dofp, "(%d, %d):Cannot allocate B, k=%d, n=%d\n", myi, myj, lda, ldb);
           fflush(dofp);
           free(A);
           return -1;
        }

        C = (double*)malloc(
                     sizeof(double)
                     *
                     (lda*ldb)
        );

        if (C == NULL)
        {
           fprintf(dofp, "(%d, %d):Cannot allocate C, m=%d, n=%d \n", myi, myj, lda, ldb);
           fflush(dofp);
           free(A);
           free(B);
           return -1;
        }

        /*
        * Initialize the matrices
        */
        InitializeMatricesLocal(
           lda, ldb, A, B, C
        );

        /*
        * Start the algorithm
        * Here n is actually the size of the matrix divided by the size of the matrix element 
        * (MSIZE/kb)
        */
        for (k = 0; k < n; k++)
        {
           int use_my_local_WB = 0;

           double *WB_l = NULL;

           /*
           * P(i,k) broadcasts a(i,k) to p(i,*) horizontally.
           * The processes holding the pivot column would be
           * (*, pivotcolproc);
           */
           int pivotcolproc = q-1;
           int cumj         = w[HMPI_RECT_INDEX(0, 0, 0, 0, p, q)];

           for (j = 1; j < q; j++)
           {
              if (k < cumj)
              {
                 pivotcolproc = j-1;
                 break;
              }
              cumj += w[HMPI_RECT_INDEX(0, j, 0, j, p, q)];
           }

           /*
           * All the processes in the process column pivotcolproc
           * horizontally send the data to other processes who share
           * common elements
           */
           if (myj == pivotcolproc)
           {
              start = MPI_Wtime();

              /*
              * Distance of the pivot column in my local matrix
              */
              int pivotcollocal = myw - cumj + k;

              /*
              * Create a temporary copy now
              */
              int tmp1   = ldb;
              int tmp2   = pivotcollocal*kb;
              for (i = 0; i < lda; i++)
              {
                 int rind = i*tmp1 + tmp2;
                 memcpy(&WA[i*kb], &A[rind], NBYTES);
              }

              /*
              * This array contains cumulative size of the messages
              * sent; 
              * In COLUMN_BASED distribution, actually only q-1 size
              * is enough
              */
              int hmsglen[q];

              for (j = 0; j < q; j++)
              {
                 hmsglen[j] = 0;
              }

              for (i = 0; i < p; i++)
              {
                 for (j = 0; j < q; j++)
                 {
                    if (j == myj)
                    {
                       continue;
                    }

                    /*
                    * Are there any matrix elements common with processes in my row?
                    */
                    int ch = h[HMPI_RECT_INDEX(myi, myj, i, j, p, q)];
   
                    if (ch == 0)
                    {
                       continue;
                    }

                    /*
                    * Send all my matrix elements common to other processes 
                    */
                    int    dest      = i*q+j;

                    /*
                    * Get to the element from where you should start sending the message
                    */
                    int elemStartHeight, startingElemInd;

                    if (i == 0)
                    {
                       startingElemInd = 0;
                       elemStartHeight = 0;
                    }
                    else
                    {
                       elemStartHeight = h[HMPI_RECT_INDEX(myi, myj, (i-1), j, p, q)];
                       hmsglen[j]     += elemStartHeight*MATELEMSIZE;
                       startingElemInd = hmsglen[j];
                    }

                    if (VERBOSE > 0)
                    {
                       fprintf(dofp, 
                          "<--COMM-->: k=%d: A communications(Send), (%d,%d) --> (%d,%d)=(%d elements) at index (%d,%d)\n", 
                          k, myi, myj, i, j, ch*MATELEMSIZE, elemStartHeight, startingElemInd);
                       fflush(dofp);
                    }

                    rc = MPI_Send(
                            &WA[startingElemInd],
                            ch*MATELEMSIZE,
                            MPI_DOUBLE,
                            dest,
                            MPI_MSG_TAG_FOR_A,
                            mxmcomm
                    );
   
                    if (rc != MPI_SUCCESS)
                    {
                       printf("k=%d, me=%d: Communications of elements of A failed\n", k, me);
                    }
                 }
              }

              end = MPI_Wtime();

              total_comm_time += (end - start);

              if (print_communication_times_to_file)
              {
                 double d1 = (lda*kb*8);
                 double d2 = 1024.;
                 double d3 = d1/d2;
                 fprintf(dofp, 
                    "<--COMM-->: k=%d: A Communications(Send), (%3f KB)=%3f seconds\n", 
                    k, d3, (end - start));
                 fflush(dofp);
              }
           }
           else
           {
              start = MPI_Wtime();

              int hmsglen = 0;

              for (i = 0; i < p; i++)
              {
                 int ch      = h[HMPI_RECT_INDEX(myi, myj, i, pivotcolproc, p, q)];
  
                 if (ch > 0)
                 {
                    int tmplen = ch*MATELEMSIZE;
                    int src    = i*q+pivotcolproc;

                    MPI_Status stat;

                    if (VERBOSE > 0)
                    {
                       fprintf(dofp, 
                           "<--COMM-->: k=%d: A communications(Receive), (%d,%d) --> (%d,%d)=(%d elements) at index %d\n", 
                           k, i, pivotcolproc, myi, myj, tmplen, hmsglen/MATELEMSIZE);
                       fflush(dofp);
                    }

                    double *WA_l = &WA[hmsglen];

                    rc = MPI_Recv(
                            WA_l,
                            tmplen,
                            MPI_DOUBLE,
                            src,
                            MPI_MSG_TAG_FOR_A,
                            mxmcomm,
                            &stat
                    );
   
                    if (rc != MPI_SUCCESS)
                    {
                       printf("k=%d, me=%d: Communications of elements of A failed\n", k, me);
                    }

                    hmsglen += tmplen;
                 }
              }

              end = MPI_Wtime();

              total_comm_time += (end - start);

              if (print_communication_times_to_file)
              {
                 double d1 = (lda*kb*8);
                 double d2 = 1024.;
                 double d3 = d1/d2;

                 fprintf(dofp, 
                    "<--COMM-->: k=%d: A Communications(Receive), (%3f KB)=%3f seconds\n", k, d3, (end - start) );
                 fflush(dofp);
              }
           }

           /*
           * P(k,j) broadcasts a(k,j) to p(*,j) vertically.
           */
           for (j = 0; j < q; j++)
           {
               if (j != myj)
               {
                  continue;
               }

               /*
               * The processes holding the pivot row would be 
               * (pivotrowproc, myj);
               */
               int pivotrowproc = p-1;
               int cumi         = h[HMPI_RECT_INDEX(0, j, 0, j, p, q)];

               for (i = 1; i < p; i++)
               {
                  if (k < cumi)
                  {
                     pivotrowproc = i-1;
                     break;
                  }
                  cumi += h[HMPI_RECT_INDEX(i, j, i, j, p, q)];
               }

               if (myi == pivotrowproc)
               {
                  start = MPI_Wtime();

                  /*
                  * Distance of the pivot row in my local matrix
                  */
                  int nelemtosend   = myw*MATELEMSIZE;
                  int pivotrowlocal = myh - cumi + k;

                  /*
                  * Get to the point from where we should start sending the data
                  * pivotrowlocal is along the column of the local matrix
                  */
                  WB_l = &B[pivotrowlocal*myw*MATELEMSIZE];
                  use_my_local_WB = 1;

                  for (i = 0; i < p; i++)
                  {
                     if (i == myi)
                     {
                        continue;
                     }

                     /*
                     * Are there any matrix elements common with processes in my process column?
                     */
                     int cw = w[HMPI_RECT_INDEX(myi, myj, i, myj, p, q)];

                     if (cw == 0)
                     {
                        continue;
                     }

                     /*
                     * Send all my matrix elements to the processes in my process column
                     */
                     int    dest      = i*q+myj;

                     if (VERBOSE > 0)
                     {
                        fprintf(dofp, 
                           "<--COMM-->: k=%d: B communications(Send), (%d,%d) --> (%d,%d)=(%d elements)\n", 
                           k, myi, myj, i, myj, nelemtosend);
                        fflush(dofp);
                     }

                     rc = MPI_Send(
                             WB_l,
                             nelemtosend,
                             MPI_DOUBLE,
                             dest,
                             MPI_MSG_TAG_FOR_B,
                             mxmcomm
                     );
    
                     if (rc != MPI_SUCCESS)
                     {
                        printf("k=%d, me=%d: Communications of elements of B failed\n", k, me);
                     }
                  }

                  end = MPI_Wtime();

                  total_comm_time += (end - start);

                  if (print_communication_times_to_file)
                  {
                     double d1 = (ldb*kb*8);
                     double d2 = 1024.;
                     double d3 = d1/d2;

                     fprintf(dofp, 
                        "<--COMM-->: k=%d: B Communications (Send), (%3f KB)=%3f seconds\n", k, d3, (end - start) );
                     fflush(dofp);
                  }
               }    
               else
               {
                  start = MPI_Wtime();

                  /*
                  * Are there any matrix elements common with process with the pivot row?
                  * For COLUMN_BASED distribution, cw = myw;
                  */
                  int cw           = w[HMPI_RECT_INDEX(myi, myj, pivotrowproc, myj, p, q)];
                  int nelemtorecv  = cw*MATELEMSIZE;
  
                  if (cw > 0)
                  {
                     int        src  = pivotrowproc*q+myj;

                     MPI_Status stat;

                     if (VERBOSE > 0)
                     {
                        fprintf(dofp, 
                           "<--COMM-->: k=%d: B communications(Receive), (%d,%d) --> (%d,%d)=(%d elements)\n", 
                           k, pivotrowproc, myj, myi, myj, nelemtorecv);
                        fflush(dofp);
                     }

                     rc = MPI_Recv(
                             WB,
                             nelemtorecv,
                             MPI_DOUBLE,
                             src,
                             MPI_MSG_TAG_FOR_B,
                             mxmcomm,
                             &stat
                     );

                     if (rc != MPI_SUCCESS)
                     {
                        printf("k=%d, me=%d: Communications of elements of B failed\n", k, me);
                     }
                  }

                  end = MPI_Wtime();

                  total_comm_time += (end - start);

                  if (print_communication_times_to_file)
                  {
                     double d1 = (ldb*kb*8);
                     double d2 = 1024.;
                     double d3 = d1/d2;

                     fprintf(dofp, 
                        "<--COMM-->: k=%d: B Communications(Receive), (%3f KB)=%3f seconds\n", k, d3, (end - start) );
                     fflush(dofp);
                  }
               }    
           }

           /*
           *  Perform local update
           */
           if (use_my_local_WB == 1)
           {
              dgemm_local_optimized(
                &lda,
                &ldb,
                &kb,
                WA,
                WB_l,
                C,
                &etime
              );
           }
           else
           {
              dgemm_local_optimized(
                &lda,
                &ldb,
                &kb,
                WA,
                WB,
                C,
                &etime
              );
           }

           total_comp_time += etime;

           if (print_computation_times_to_file)
           {
              fprintf(dofp, "<**COMP**>: k=%d: (m, n, k)=(%d,%d,%d), %3f seconds\n", k, lda, ldb, kb, etime);
              fflush(dofp);
           }
        }

        if (CHECK_RESULTS > 0)
        {
           int min_lda_ldb = (lda < ldb) ? lda : ldb;
           fprintf(dofp, "C[]:\n");
           for (i = 0; i < min_lda_ldb; i++)
           {
              fprintf(dofp, "%3f ", C[i*ldb+i]);
           }
           fprintf(dofp, "\n");
           fflush(dofp);
        }

        /*
        * Free the matrices
        */
        free(WA);
        free(WB);
        free(A);
        free(B);
        free(C);

        return 0;
    }

    /*-----------------------------------------------------------------------------------------*/

