
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pcscal_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int N          =  params[0];
   int IX         =  params[1+1+1];
   int JX         =  params[1+1+1+1];
   int *DESCX     = &params[1+1+1+1+1];
   int INCX       =  params[1+1+1+1+1+DLEN_];

   hscalcomplex ALPHA, *x;

   /*
   * Optimal process arrangement is usually 1xp
   * Use this heuristic to allocate matrix A and
   * vectors x and y of appropriate sizes
   */
   int num_of_computers = HMPI_Get_number_of_computers();
   int reconn = (((N/num_of_computers) >= 1) ? (N/num_of_computers) : 1);
   int RECONX = ( 1 + ( reconn - 1 )*abs( INCX ) );

   ALPHA.r        = params[1];
   ALPHA.i        = params[1+1];

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "N %d reconn %d ALPHA (%f, %f) IX %d JX %d INCX %d",
           N, reconn, ALPHA.r, ALPHA.i, IX, JX, INCX
      );
   }

   //
   // Allocate the vector x
   x = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 RECONX
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the vector x
   for (i = 0; i < RECONX; i++)
   {
      x[i].r = 2.;
      x[i].i = 2.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS cscal routine
   hscal_cscal_benchmark(
     &reconn,
     &ALPHA,
     x,
     &INCX
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the vector x
   free(x);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pcscal_modelp_noauto(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG];

   if (ALPHA[REAL_PART] == 0.f)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.f)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.f)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.f)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   *model_count = 1        /* N */
                  +
                  2        /* ALPHA */
                  +
                  1        /* JX */        
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count), 
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                               = *N;
   (*model_params)[1]                               = alphai[REAL_PART];
   (*model_params)[1+1]                             = alphai[IMAG_PART];
   (*model_params)[1+1+1]                           = *IX;
   (*model_params)[1+1+1+1]                         = *JX;
   (*model_params)[1+1+1+1+1+DLEN_]                 = *INCX;
   (*model_params)[1+1+1+1+1+DLEN_+1]               = *nprow;
   (*model_params)[1+1+1+1+1+DLEN_+1+1]             = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                 = DESCX[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pcscal_modelp(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG];

   if (ALPHA[REAL_PART] == 0.f)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.f)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.f)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.f)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   *model_count = 1        /* N */
                  +
                  2        /* ALPHA */
                  +
                  1        /* JX */        
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
   ;

   *model_params = (int*)malloc(
                         sizeof(int)
                         *
                         (*model_count)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                               = *N;
   (*model_params)[1]                               = alphai[REAL_PART];
   (*model_params)[1+1]                             = alphai[IMAG_PART];
   (*model_params)[1+1+1]                           = *IX;
   (*model_params)[1+1+1+1]                         = *JX;
   (*model_params)[1+1+1+1+1+DLEN_]                 = *INCX;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                 = DESCX[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcscal_recon(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pcscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      {
         rc = HMPI_Recon_t(
                  &hscal_pcscal_benchmark_function,
                  model_params,
                  model_count,
                  &output_p
         );

         if (rc != HMPI_SUCCESS)
         {
            free(model_params);
            return rc;
         }
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pcscal_timeof(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pcscal_modelp_noauto(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pcscal_group_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcscal_modelp_noauto(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcscal_group_auto_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcscal_group_heuristic_pauto_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pcscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcscal_group_heuristic_auto_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pzscal_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int N          =  params[0];
   int IX         =  params[1+1+1];
   int JX         =  params[1+1+1+1];
   int *DESCX     = &params[1+1+1+1+1];
   int INCX       =  params[1+1+1+1+1+DLEN_];

   hscaldoublecomplex ALPHA, *x;

   /*
   * Optimal process arrangement is usually 1xp
   * Use this heuristic to allocate matrix A and
   * vectors x and y of appropriate sizes
   */
   int num_of_computers = HMPI_Get_number_of_computers();
   int reconn = (((N/num_of_computers) >= 1) ? (N/num_of_computers) : 1);
   int RECONX = ( 1 + ( reconn - 1 )*abs( INCX ) );

   ALPHA.r        = params[1];
   ALPHA.i        = params[1+1];

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "N %d reconn %d ALPHA (%f, %f) IX %d JX %d INCX %d",
           N, reconn, ALPHA.r, ALPHA.i, IX, JX, INCX
      );
   }

   //
   // Allocate the vector x
   x = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 RECONX
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the vector x
   for (i = 0; i < RECONX; i++)
   {
      x[i].r = 2.;
      x[i].i = 2.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS zscal routine
   hscal_zscal_benchmark(
     &reconn,
     &ALPHA,
     x,
     &INCX
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the vector x
   free(x);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pzscal_modelp_noauto(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha
   */
   int alphai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   *model_count = 1        /* N */
                  +
                  2        /* ALPHA */
                  +
                  1        /* JX */        
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count), 
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                               = *N;
   (*model_params)[1]                               = alphai[REAL_PART];
   (*model_params)[1+1]                             = alphai[IMAG_PART];
   (*model_params)[1+1+1]                           = *IX;
   (*model_params)[1+1+1+1]                         = *JX;
   (*model_params)[1+1+1+1+1+DLEN_]                 = *INCX;
   (*model_params)[1+1+1+1+1+DLEN_+1]               = *nprow;
   (*model_params)[1+1+1+1+1+DLEN_+1+1]             = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                 = DESCX[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pzscal_modelp(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha
   */
   int alphai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   *model_count = 1        /* N */
                  +
                  2        /* ALPHA */
                  +
                  1        /* JX */        
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
   ;

   *model_params = (int*)malloc(
                         sizeof(int)
                         *
                         (*model_count)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                               = *N;
   (*model_params)[1]                               = alphai[REAL_PART];
   (*model_params)[1+1]                             = alphai[IMAG_PART];
   (*model_params)[1+1+1]                           = *IX;
   (*model_params)[1+1+1+1]                         = *JX;
   (*model_params)[1+1+1+1+1+DLEN_]                 = *INCX;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                 = DESCX[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzscal_recon(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pzscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      {
         rc = HMPI_Recon_t(
                  &hscal_pzscal_benchmark_function,
                  model_params,
                  model_count,
                  &output_p
         );

         if (rc != HMPI_SUCCESS)
         {
            free(model_params);
            return rc;
         }
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pzscal_timeof(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pzscal_modelp_noauto(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pzscal_group_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzscal_modelp_noauto(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzscal_group_auto_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzscal_group_heuristic_pauto_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pzscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzscal_group_heuristic_auto_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pdscal_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int N          =  params[0];
   double ALPHA   =  params[1];
   int IX         =  params[1+1];
   int JX         =  params[1+1+1];
   int *DESCX     = &params[1+1+1+1];
   int INCX       =  params[1+1+1+1+DLEN_];

   hscaldoublereal *x;

   /*
   * Optimal process arrangement is usually 1xp
   * Use this heuristic to allocate matrix A and
   * vectors x and y of appropriate sizes
   */
   int num_of_computers = HMPI_Get_number_of_computers();
   int reconn = (((N/num_of_computers) >= 1) ? (N/num_of_computers) : 1);
   int RECONX = ( 1 + ( reconn - 1 )*abs( INCX ) );

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "N %d reconn %d ALPHA %f IX %d JX %d INCX %d",
           N, reconn, ALPHA, IX, JX, INCX
      );
   }

   //
   // Allocate the vector x
   x = (hscaldoublereal*)malloc(
                 sizeof(hscaldoublereal)
                 *
                 RECONX
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the vector x
   for (i = 0; i < RECONX; i++)
   {
      x[i] = 2.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS dscal routine
   hscal_dscal_benchmark(
     &reconn,
     &ALPHA,
     x,
     &INCX
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the vector x
   free(x);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdscal_modelp_noauto(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* JX */        
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count), 
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                               = *N;
   (*model_params)[1]                               = *ALPHA;
   (*model_params)[1+1]                             = *IX;
   (*model_params)[1+1+1]                           = *JX;
   (*model_params)[1+1+1+1+DLEN_]                   = *INCX;
   (*model_params)[1+1+1+1+DLEN_+1]                 = *nprow;
   (*model_params)[1+1+1+1+DLEN_+1+1]               = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+i]                   = DESCX[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdscal_modelp(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* JX */        
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
   ;

   *model_params = (int*)malloc(
                         sizeof(int)
                         *
                         (*model_count)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                               = *N;
   (*model_params)[1]                               = *ALPHA;
   (*model_params)[1+1]                             = *IX;
   (*model_params)[1+1+1]                           = *JX;
   (*model_params)[1+1+1+1+DLEN_]                   = *INCX;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+i]                   = DESCX[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdscal_recon(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pdscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      {
         rc = HMPI_Recon_t(
                  &hscal_pdscal_benchmark_function,
                  model_params,
                  model_count,
                  &output_p
         );

         if (rc != HMPI_SUCCESS)
         {
            free(model_params);
            return rc;
         }
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pdscal_timeof(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdscal_modelp_noauto(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pdscal_group_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdscal_modelp_noauto(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdscal_group_auto_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdscal_group_heuristic_pauto_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdscal_group_heuristic_auto_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_psscal_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int N          =  params[0];
   float ALPHA    =  params[1];
   int IX         =  params[1+1];
   int JX         =  params[1+1+1];
   int *DESCX     = &params[1+1+1+1];
   int INCX       =  params[1+1+1+1+DLEN_];

   hscalreal *x;

   /*
   * Optimal process arrangement is usually 1xp
   * Use this heuristic to allocate matrix A and
   * vectors x and y of appropriate sizes
   */
   int num_of_computers = HMPI_Get_number_of_computers();
   int reconn = (((N/num_of_computers) >= 1) ? (N/num_of_computers) : 1);
   int RECONX = ( 1 + ( reconn - 1 )*abs( INCX ) );

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "N %d reconn %d ALPHA %f IX %d JX %d INCX %d",
           N, reconn, ALPHA, IX, JX, INCX
      );
   }

   //
   // Allocate the vector x
   x = (hscalreal*)malloc(
                 sizeof(hscalreal)
                 *
                 RECONX
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the vector x
   for (i = 0; i < RECONX; i++)
   {
      x[i] = 2.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS sscal routine
   hscal_sscal_benchmark(
     &reconn,
     &ALPHA,
     x,
     &INCX
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the vector x
   free(x);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_psscal_modelp_noauto(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* JX */        
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count), 
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                               = *N;
   (*model_params)[1]                               = *ALPHA;
   (*model_params)[1+1]                             = *IX;
   (*model_params)[1+1+1]                           = *JX;
   (*model_params)[1+1+1+1+DLEN_]                   = *INCX;
   (*model_params)[1+1+1+1+DLEN_+1]                 = *nprow;
   (*model_params)[1+1+1+1+DLEN_+1+1]               = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+i]                   = DESCX[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_psscal_modelp(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* JX */        
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
   ;

   *model_params = (int*)malloc(
                         sizeof(int)
                         *
                         (*model_count)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                               = *N;
   (*model_params)[1]                               = *ALPHA;
   (*model_params)[1+1]                             = *IX;
   (*model_params)[1+1+1]                           = *JX;
   (*model_params)[1+1+1+1+DLEN_]                   = *INCX;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+i]                   = DESCX[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psscal_recon(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_psscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      {
         rc = HMPI_Recon_t(
                  &hscal_psscal_benchmark_function,
                  model_params,
                  model_count,
                  &output_p
         );

         if (rc != HMPI_SUCCESS)
         {
            free(model_params);
            return rc;
         }
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_psscal_timeof(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_psscal_modelp_noauto(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_psscal_group_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psscal_modelp_noauto(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psscal_group_auto_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psscal_group_heuristic_pauto_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_psscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psscal_group_heuristic_auto_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pzdscal_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int N          =  params[0];
   double ALPHA   =  params[1];
   int IX         =  params[1+1];
   int JX         =  params[1+1+1];
   int *DESCX     = &params[1+1+1+1];
   int INCX       =  params[1+1+1+1+DLEN_];

   hscaldoublecomplex *x;

   /*
   * Optimal process arrangement is usually 1xp/px1
   * Use this heuristic to allocate matrix A and
   * vectors x and y of appropriate sizes
   */
   int num_of_computers = HMPI_Get_number_of_computers();
   int reconn = (((N/num_of_computers) >= 1) ? (N/num_of_computers) : 1);
   int RECONX = ( 1 + ( reconn - 1 )*abs( INCX ) );

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "N %d reconn %d ALPHA %f IX %d JX %d INCX %d",
           N, reconn, ALPHA, IX, JX, INCX
      );
   }

   //
   // Allocate the vector x
   x = (hscaldoublecomplex*)malloc(
                       sizeof(hscaldoublecomplex)
                       *
                       RECONX
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the vector x
   for (i = 0; i < RECONX; i++)
   {
      x[i].r = 2.;
      x[i].i = 2.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS zdscal routine
   hscal_zdscal_benchmark(
     &reconn,
     &ALPHA,
     x,
     &INCX
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the vector x
   free(x);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pzdscal_modelp_noauto(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* JX */        
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count), 
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                               = *N;
   (*model_params)[1]                               = *ALPHA;
   (*model_params)[1+1]                             = *IX;
   (*model_params)[1+1+1]                           = *JX;
   (*model_params)[1+1+1+1+DLEN_]                   = *INCX;
   (*model_params)[1+1+1+1+DLEN_+1]                 = *nprow;
   (*model_params)[1+1+1+1+DLEN_+1+1]               = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+i]                   = DESCX[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pzdscal_modelp(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* JX */        
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
   ;

   *model_params = (int*)malloc(
                         sizeof(int)
                         *
                         (*model_count)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                               = *N;
   (*model_params)[1]                               = *ALPHA;
   (*model_params)[1+1]                             = *IX;
   (*model_params)[1+1+1]                           = *JX;
   (*model_params)[1+1+1+1+DLEN_]                   = *INCX;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+i]                   = DESCX[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzdscal_recon(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pzdscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      {
         rc = HMPI_Recon_t(
                  &hscal_pdscal_benchmark_function,
                  model_params,
                  model_count,
                  &output_p
         );

         if (rc != HMPI_SUCCESS)
         {
            free(model_params);
            return rc;
         }
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pzdscal_timeof(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pzdscal_modelp_noauto(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pzdscal_group_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzdscal_modelp_noauto(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzdscal_group_auto_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzdscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzdscal_group_heuristic_pauto_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pzdscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzdscal_group_heuristic_auto_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzdscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pcsscal_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int N          =  params[0];
   float ALPHA    =  params[1];
   int IX         =  params[1+1];
   int JX         =  params[1+1+1];
   int *DESCX     = &params[1+1+1+1];
   int INCX       =  params[1+1+1+1+DLEN_];

   hscalcomplex *x;

   /*
   * Optimal process arrangement is usually 1xp
   * Use this heuristic to allocate matrix A and
   * vectors x and y of appropriate sizes
   */
   int num_of_computers = HMPI_Get_number_of_computers();
   int reconn = (((N/num_of_computers) >= 1) ? (N/num_of_computers) : 1);
   int RECONX = ( 1 + ( reconn - 1 )*abs( INCX ) );

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "N %d reconn %d ALPHA %f IX %d JX %d INCX %d",
           N, reconn, ALPHA, IX, JX, INCX
      );
   }

   //
   // Allocate the vector x
   x = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 RECONX
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the vector x
   for (i = 0; i < RECONX; i++)
   {
      x[i].i = 2.f;
      x[i].r = 2.f;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS csscal routine
   hscal_csscal_benchmark(
     &reconn,
     &ALPHA,
     x,
     &INCX
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the vector x
   free(x);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pcsscal_modelp_noauto(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* JX */        
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count), 
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                               = *N;
   (*model_params)[1]                               = *ALPHA;
   (*model_params)[1+1]                             = *IX;
   (*model_params)[1+1+1]                           = *JX;
   (*model_params)[1+1+1+1+DLEN_]                   = *INCX;
   (*model_params)[1+1+1+1+DLEN_+1]                 = *nprow;
   (*model_params)[1+1+1+1+DLEN_+1+1]               = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+i]                   = DESCX[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pcsscal_modelp(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* JX */        
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
   ;

   *model_params = (int*)malloc(
                         sizeof(int)
                         *
                         (*model_count)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                               = *N;
   (*model_params)[1]                               = *ALPHA;
   (*model_params)[1+1]                             = *IX;
   (*model_params)[1+1+1]                           = *JX;
   (*model_params)[1+1+1+1+DLEN_]                   = *INCX;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+i]                   = DESCX[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcsscal_recon(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pcsscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      {
         rc = HMPI_Recon_t(
                  &hscal_psscal_benchmark_function,
                  model_params,
                  model_count,
                  &output_p
         );

         if (rc != HMPI_SUCCESS)
         {
            free(model_params);
            return rc;
         }
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pcsscal_timeof(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pcsscal_modelp_noauto(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pcsscal_group_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcsscal_modelp_noauto(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcsscal_group_auto_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcsscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcsscal_group_heuristic_pauto_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pcsscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcsscal_group_heuristic_auto_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcsscal_modelp(
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

