
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pcswap_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int N          =  params[0];
   int IX         =  params[1];
   int JX         =  params[1+1];
   int *DESCX     = &params[1+1+1];
   int INCX       =  params[1+1+1+DLEN_];
   int IY         =  params[1+1+1+DLEN_+1];
   int JY         =  params[1+1+1+DLEN_+1+1];
   int *DESCY     = &params[1+1+1+DLEN_+1+1+1];
   int INCY       =  params[1+1+1+DLEN_+1+1+1+DLEN_];

   hscalcomplex *x, *y;

   /*
   * Optimal process arrangement is usually 1xp
   * Use this heuristic to allocate matrix A and
   * vectors x and y of appropriate sizes
   */
   int num_of_computers = HMPI_Get_number_of_computers();
   int reconn = (((N/num_of_computers) >= 1) ? (N/num_of_computers) : 1);
   int RECONX = ( 1 + ( reconn - 1 )*abs( INCX ) );
   int RECONY = ( 1 + ( reconn - 1 )*abs( INCY ) );

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "N %d reconn %d IX %d JX %d "
           "INCX %d IY %d JY %d INCY %d",
           N, reconn, IX, JX,
           INCX, IY, JY, INCY
      );
   }

   //
   // Allocate the vectors x and y
   x = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 RECONX
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   y = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 RECONY
   );

   if (y == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the vectors x and y
   for (i = 0; i < RECONX; i++)
   {
      x[i].r = 2.;
      x[i].i = 2.;
   }

   for (i = 0; i < RECONY; i++)
   {
      y[i].r = 1.;
      y[i].i = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS cswap routine
   hscal_cswap_benchmark(
     &reconn,
     x,
     &INCX,
     y,
     &INCY
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the vectors x and y
   free(x);
   free(y);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pcswap_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* N */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count), 
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                               = *N;
   (*model_params)[1]                               = *IX;
   (*model_params)[1+1]                             = *JX;
   (*model_params)[1+1+1+DLEN_]                     = *INCX;
   (*model_params)[1+1+1+DLEN_+1]                   = *IY;
   (*model_params)[1+1+1+DLEN_+1+1]                 = *JY;
   (*model_params)[1+1+1+DLEN_+1+1+1+DLEN_]         = *INCY;
   (*model_params)[1+1+1+DLEN_+1+1+1+DLEN_+1]       = *nprow;
   (*model_params)[1+1+1+DLEN_+1+1+1+DLEN_+1+1]     = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+i]                     = DESCX[i];
       (*model_params)[1+1+1+DLEN_+1+1+1+i]         = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pcswap_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* N */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
   ;

   *model_params = (int*)malloc(
                         sizeof(int)
                         *
                         (*model_count)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                               = *N;
   (*model_params)[1]                               = *IX;
   (*model_params)[1+1]                             = *JX;
   (*model_params)[1+1+1+DLEN_]                     = *INCX;
   (*model_params)[1+1+1+DLEN_+1]                   = *IY;
   (*model_params)[1+1+1+DLEN_+1+1]                 = *JY;
   (*model_params)[1+1+1+DLEN_+1+1+1+DLEN_]         = *INCY;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+i]                     = DESCX[i];
       (*model_params)[1+1+1+DLEN_+1+1+1+i]         = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcswap_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pcswap_modelp(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      {
         rc = HMPI_Recon_t(
                  &hscal_pcswap_benchmark_function,
                  model_params,
                  model_count,
                  &output_p
         );

         if (rc != HMPI_SUCCESS)
         {
            free(model_params);
            return rc;
         }
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pcswap_timeof(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pcswap_modelp_noauto(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pcswap_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcswap_modelp_noauto(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcswap_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcswap_modelp(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcswap_group_heuristic_pauto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pcswap_modelp(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcswap_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcswap_modelp(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pdswap_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int N          =  params[0];
   int IX         =  params[1];
   int JX         =  params[1+1];
   int *DESCX     = &params[1+1+1];
   int INCX       =  params[1+1+1+DLEN_];
   int IY         =  params[1+1+1+DLEN_+1];
   int JY         =  params[1+1+1+DLEN_+1+1];
   int *DESCY     = &params[1+1+1+DLEN_+1+1+1];
   int INCY       =  params[1+1+1+DLEN_+1+1+1+DLEN_];

   hscaldoublereal *x, *y;

   /*
   * Optimal process arrangement is usually 1xp
   * Use this heuristic to allocate matrix A and
   * vectors x and y of appropriate sizes
   */
   int num_of_computers = HMPI_Get_number_of_computers();
   int reconn = (((N/num_of_computers) >= 1) ? (N/num_of_computers) : 1);
   int RECONX = ( 1 + ( reconn - 1 )*abs( INCX ) );
   int RECONY = ( 1 + ( reconn - 1 )*abs( INCY ) );

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "N %d reconn %d IX %d JX %d "
           "INCX %d IY %d JY %d INCY %d",
           N, reconn, IX, JX,
           INCX, IY, JY, INCY
      );
   }

   //
   // Allocate the vectors x and y
   x = (hscaldoublereal*)malloc(
                 sizeof(hscaldoublereal)
                 *
                 RECONX
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   y = (hscaldoublereal*)malloc(
                 sizeof(hscaldoublereal)
                 *
                 RECONY
   );

   if (y == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the vectors x and y
   for (i = 0; i < RECONX; i++)
   {
      x[i] = 2.;
   }

   for (i = 0; i < RECONY; i++)
   {
      y[i] = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS dswap routine
   hscal_dswap_benchmark(
     &reconn,
     x,
     &INCX,
     y,
     &INCY
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the vectors x and y
   free(x);
   free(y);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdswap_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* N */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count), 
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                               = *N;
   (*model_params)[1]                               = *IX;
   (*model_params)[1+1]                             = *JX;
   (*model_params)[1+1+1+DLEN_]                     = *INCX;
   (*model_params)[1+1+1+DLEN_+1]                   = *IY;
   (*model_params)[1+1+1+DLEN_+1+1]                 = *JY;
   (*model_params)[1+1+1+DLEN_+1+1+1+DLEN_]         = *INCY;
   (*model_params)[1+1+1+DLEN_+1+1+1+DLEN_+1]       = *nprow;
   (*model_params)[1+1+1+DLEN_+1+1+1+DLEN_+1+1]     = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+i]                     = DESCX[i];
       (*model_params)[1+1+1+DLEN_+1+1+1+i]         = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdswap_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* N */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
   ;

   *model_params = (int*)malloc(
                         sizeof(int)
                         *
                         (*model_count)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                               = *N;
   (*model_params)[1]                               = *IX;
   (*model_params)[1+1]                             = *JX;
   (*model_params)[1+1+1+DLEN_]                     = *INCX;
   (*model_params)[1+1+1+DLEN_+1]                   = *IY;
   (*model_params)[1+1+1+DLEN_+1+1]                 = *JY;
   (*model_params)[1+1+1+DLEN_+1+1+1+DLEN_]         = *INCY;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+i]                     = DESCX[i];
       (*model_params)[1+1+1+DLEN_+1+1+1+i]         = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdswap_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pdswap_modelp(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      {
         rc = HMPI_Recon_t(
                  &hscal_pdswap_benchmark_function,
                  model_params,
                  model_count,
                  &output_p
         );

         if (rc != HMPI_SUCCESS)
         {
            free(model_params);
            return rc;
         }
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pdswap_timeof(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdswap_modelp_noauto(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pdswap_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdswap_modelp_noauto(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdswap_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdswap_modelp(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdswap_group_heuristic_pauto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdswap_modelp(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdswap_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdswap_modelp(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_psswap_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int N          =  params[0];
   int IX         =  params[1];
   int JX         =  params[1+1];
   int *DESCX     = &params[1+1+1];
   int INCX       =  params[1+1+1+DLEN_];
   int IY         =  params[1+1+1+DLEN_+1];
   int JY         =  params[1+1+1+DLEN_+1+1];
   int *DESCY     = &params[1+1+1+DLEN_+1+1+1];
   int INCY       =  params[1+1+1+DLEN_+1+1+1+DLEN_];

   hscalreal *x, *y;

   /*
   * Optimal process arrangement is usually 1xp
   * Use this heuristic to allocate matrix A and
   * vectors x and y of appropriate sizes
   */
   int num_of_computers = HMPI_Get_number_of_computers();
   int reconn = (((N/num_of_computers) >= 1) ? (N/num_of_computers) : 1);
   int RECONX = ( 1 + ( reconn - 1 )*abs( INCX ) );
   int RECONY = ( 1 + ( reconn - 1 )*abs( INCY ) );

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "N %d reconn %d IX %d JX %d "
           "INCX %d IY %d JY %d INCY %d",
           N, reconn, IX, JX,
           INCX, IY, JY, INCY
      );
   }

   //
   // Allocate the vectors x and y
   x = (hscalreal*)malloc(
                 sizeof(hscalreal)
                 *
                 RECONX
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   y = (hscalreal*)malloc(
                 sizeof(hscalreal)
                 *
                 RECONY
   );

   if (y == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the vectors x and y
   for (i = 0; i < RECONX; i++)
   {
      x[i] = 2.;
   }

   for (i = 0; i < RECONY; i++)
   {
      y[i] = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS sswap routine
   hscal_sswap_benchmark(
     &reconn,
     x,
     &INCX,
     y,
     &INCY
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the vectors x and y
   free(x);
   free(y);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_psswap_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* N */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count), 
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                               = *N;
   (*model_params)[1]                               = *IX;
   (*model_params)[1+1]                             = *JX;
   (*model_params)[1+1+1+DLEN_]                     = *INCX;
   (*model_params)[1+1+1+DLEN_+1]                   = *IY;
   (*model_params)[1+1+1+DLEN_+1+1]                 = *JY;
   (*model_params)[1+1+1+DLEN_+1+1+1+DLEN_]         = *INCY;
   (*model_params)[1+1+1+DLEN_+1+1+1+DLEN_+1]       = *nprow;
   (*model_params)[1+1+1+DLEN_+1+1+1+DLEN_+1+1]     = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+i]                     = DESCX[i];
       (*model_params)[1+1+1+DLEN_+1+1+1+i]         = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_psswap_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* N */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
   ;

   *model_params = (int*)malloc(
                         sizeof(int)
                         *
                         (*model_count)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                               = *N;
   (*model_params)[1]                               = *IX;
   (*model_params)[1+1]                             = *JX;
   (*model_params)[1+1+1+DLEN_]                     = *INCX;
   (*model_params)[1+1+1+DLEN_+1]                   = *IY;
   (*model_params)[1+1+1+DLEN_+1+1]                 = *JY;
   (*model_params)[1+1+1+DLEN_+1+1+1+DLEN_]         = *INCY;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+i]                     = DESCX[i];
       (*model_params)[1+1+1+DLEN_+1+1+1+i]         = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psswap_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_psswap_modelp(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      {
         rc = HMPI_Recon_t(
                  &hscal_psswap_benchmark_function,
                  model_params,
                  model_count,
                  &output_p
         );

         if (rc != HMPI_SUCCESS)
         {
            free(model_params);
            return rc;
         }
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_psswap_timeof(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_psswap_modelp_noauto(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_psswap_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psswap_modelp_noauto(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psswap_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psswap_modelp(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psswap_group_heuristic_pauto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_psswap_modelp(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psswap_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psswap_modelp(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/


double hscal_pzswap_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int N          =  params[0];
   int IX         =  params[1];
   int JX         =  params[1+1];
   int *DESCX     = &params[1+1+1];
   int INCX       =  params[1+1+1+DLEN_];
   int IY         =  params[1+1+1+DLEN_+1];
   int JY         =  params[1+1+1+DLEN_+1+1];
   int *DESCY     = &params[1+1+1+DLEN_+1+1+1];
   int INCY       =  params[1+1+1+DLEN_+1+1+1+DLEN_];

   hscaldoublecomplex *x, *y;

   /*
   * Optimal process arrangement is usually 1xp
   * Use this heuristic to allocate matrix A and
   * vectors x and y of appropriate sizes
   */
   int num_of_computers = HMPI_Get_number_of_computers();
   int reconn = (((N/num_of_computers) >= 1) ? (N/num_of_computers) : 1);
   int RECONX = ( 1 + ( reconn - 1 )*abs( INCX ) );
   int RECONY = ( 1 + ( reconn - 1 )*abs( INCY ) );

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "N %d reconn %d IX %d JX %d "
           "INCX %d IY %d JY %d INCY %d",
           N, reconn, IX, JX,
           INCX, IY, JY, INCY
      );
   }

   //
   // Allocate the vectors x and y
   x = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 RECONX
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   y = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 RECONY
   );

   if (y == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the vectors x and y
   for (i = 0; i < RECONX; i++)
   {
      x[i].r = 2.;
      x[i].i = 2.;
   }

   for (i = 0; i < RECONY; i++)
   {
      y[i].r = 1.;
      y[i].i = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS zswap routine
   hscal_zswap_benchmark(
     &reconn,
     x,
     &INCX,
     y,
     &INCY
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the vectors x and y
   free(x);
   free(y);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pzswap_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* N */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count), 
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                               = *N;
   (*model_params)[1]                               = *IX;
   (*model_params)[1+1]                             = *JX;
   (*model_params)[1+1+1+DLEN_]                     = *INCX;
   (*model_params)[1+1+1+DLEN_+1]                   = *IY;
   (*model_params)[1+1+1+DLEN_+1+1]                 = *JY;
   (*model_params)[1+1+1+DLEN_+1+1+1+DLEN_]         = *INCY;
   (*model_params)[1+1+1+DLEN_+1+1+1+DLEN_+1]       = *nprow;
   (*model_params)[1+1+1+DLEN_+1+1+1+DLEN_+1+1]     = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+i]                     = DESCX[i];
       (*model_params)[1+1+1+DLEN_+1+1+1+i]         = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pzswap_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* N */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
   ;

   *model_params = (int*)malloc(
                         sizeof(int)
                         *
                         (*model_count)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                               = *N;
   (*model_params)[1]                               = *IX;
   (*model_params)[1+1]                             = *JX;
   (*model_params)[1+1+1+DLEN_]                     = *INCX;
   (*model_params)[1+1+1+DLEN_+1]                   = *IY;
   (*model_params)[1+1+1+DLEN_+1+1]                 = *JY;
   (*model_params)[1+1+1+DLEN_+1+1+1+DLEN_]         = *INCY;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+i]                     = DESCX[i];
       (*model_params)[1+1+1+DLEN_+1+1+1+i]         = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzswap_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pzswap_modelp(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      {
         rc = HMPI_Recon_t(
                  &hscal_pzswap_benchmark_function,
                  model_params,
                  model_count,
                  &output_p
         );

         if (rc != HMPI_SUCCESS)
         {
            free(model_params);
            return rc;
         }
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pzswap_timeof(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pzswap_modelp_noauto(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pzswap_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzswap_modelp_noauto(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzswap_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzswap_modelp(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzswap_group_heuristic_pauto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pzswap_modelp(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzswap_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzswap_modelp(
                N,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
