
#include "../../include/hscalapack.h"

/*  ===================================================================== */
/*  Purpose */
/*  ======= */
/*  The utility of these functions is to return the algorithmic blocking factor */
/*  used in ScaLAPACK */
/*  These functions are used in the performance models */
/*  To decrease the execution times of the performance models */
/*  but at the same time to keep the accuracy of the predictions */
/*  we must use a value depending on the problem size */

/*  This function returns the algorithmic blocking factor and not the data */
/*  distribution blocking factor */

/*  Arguments */
/*  ========= */

/*  PREC    (global input) CHARACTER*1 */
/*          On input, PREC specifies the precision for which the  logical */
/*          block size should be returned as follows: */
/*             PREC = 'S' or 's' single precision real, */
/*             PREC = 'D' or 'd' double precision real, */
/*             PREC = 'C' or 'c' single precision complex, */
/*             PREC = 'Z' or 'z' double precision complex, */
/*             PREC = 'I' or 'i' integer. */

/*  Notes */
/*  ===== */

/*  Before modifying this routine, be aware of the following: */

/*  1) The value this function returns must be STRICTLY LARGER THAN ZERO, */

/*  2) If you are planning to link your program with different  instances */
/*  of the library, (for example on a heterogeneous  machine),  you  MUST */
/*  compile each instance of the library with the  EXACT SAME  version of */
/*  this routine for obvious inter-operability reasons. */

/*  ===================================================================== */

/*------------------------------------------------------------------*/

int hscal_pilaenv
(
    char *precision
)
{
    /*
    * Returns the value used by ScaLAPACK
    */
    int ictxt = -1;
    int scalapack_pilaenv = pilaenv_(
                            &ictxt,
                            precision
    );

    return scalapack_pilaenv;

/*     End of hscal_pilaenv */

}

/*------------------------------------------------------------------*/

int hscal_pilaenv_n
(
    const int N,
    char *precision
)
{
    /*
    * Determine a value less than the problem size but a multiple of the 
    * value used in the ScaLAPACK library
    */
    int hscalapack_pilaenv;
    int ictxt = -1;
    int scalapack_pilaenv = pilaenv_(
                            &ictxt,
                            precision
    );

    if (scalapack_pilaenv >= N)
    {
       hscalapack_pilaenv = scalapack_pilaenv;
    }
    else if (scalapack_pilaenv >= (N/2))
    {
       hscalapack_pilaenv = scalapack_pilaenv;
    }
    else
    {
       int r = (N/2)/scalapack_pilaenv;
       hscalapack_pilaenv = r*scalapack_pilaenv;
    }

    return hscalapack_pilaenv;

/*     End of hscal_pilaenv_n */

}

/*------------------------------------------------------------------*/

int hscal_pilaenv_m_n
(
    const int M,
    const int N,
    char *precision
)
{
    /*
    * Determine a value less than the problem size but a multiple of the 
    * value used in the ScaLAPACK library
    */
    int min_M_N = ((M <= N) ? M : N);
    int hscalapack_pilaenv;
    int ictxt = -1;
    int scalapack_pilaenv = pilaenv_(
                            &ictxt,
                            precision
    );

    if (scalapack_pilaenv >= min_M_N)
    {
       hscalapack_pilaenv = scalapack_pilaenv;
    }
    else if (scalapack_pilaenv >= (min_M_N/2))
    {
       hscalapack_pilaenv = scalapack_pilaenv;
    }
    else
    {
       int r = (min_M_N/2)/scalapack_pilaenv;
       hscalapack_pilaenv = r*scalapack_pilaenv;
    }

    return hscalapack_pilaenv;

/*     End of hscal_pilaenv_m_n */

}

/*------------------------------------------------------------------*/

int hscal_pilaenv_m_n_k
(
    const int M,
    const int N,
    const int K,
    char *precision
)
{
    /*
    * Determine a value less than the problem size but a multiple of the 
    * value used in the ScaLAPACK library
    */
    int min_M_N = ((M <= N) ? M : N);
    int min_M_N_K = ((min_M_N <= K) ? min_M_N : K);
    int hscalapack_pilaenv;
    int ictxt = -1;
    int scalapack_pilaenv = pilaenv_(
                            &ictxt,
                            precision
    );

    if (scalapack_pilaenv >= min_M_N_K)
    {
       hscalapack_pilaenv = scalapack_pilaenv;
    }
    else if (scalapack_pilaenv >= (min_M_N_K/2))
    {
       hscalapack_pilaenv = scalapack_pilaenv;
    }
    else
    {
       int r = (min_M_N_K/2)/scalapack_pilaenv;
       hscalapack_pilaenv = r*scalapack_pilaenv;
    }

    return hscalapack_pilaenv;

/*     End of hscal_pilaenv_m_n_k */

}

/*------------------------------------------------------------------*/

