
#include "hscal_hashtab.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

static const unsigned int primes[] = {
53, 97, 193, 389,
769, 1543, 3079, 6151,
12289, 24593, 49157, 98317,
196613, 393241, 786433, 1572869,
3145739, 6291469, 12582917, 25165843,
50331653, 100663319, 201326611, 402653189,
805306457, 1610612741
};

const unsigned int prime_table_length = sizeof(primes)/sizeof(primes[0]);
const float max_load_factor = 0.65;

/*---------------------------------------------------------------------------*/

unsigned int hscal_indexFor(
    unsigned int tablelength,
    unsigned int hashvalue
)
{
    return (hashvalue % tablelength);
}

/*---------------------------------------------------------------------------*/

hscal_hashtable *
hscal_create_hashtable(
     unsigned int minsize,
     unsigned int (*hashf) (void*),
     int (*eqf) (void*,void*)
)
{
    hscal_hashtable *h;
    unsigned int pindex, size = primes[0];

    /* 
    * Check requested hashtable isn't too large 
    */
    if (minsize > (1u << 30)) 
    {
       return NULL;
    }

    /* 
    * Enforce size as prime 
    */
    for (pindex=0; pindex < prime_table_length; pindex++) 
    {
        if (primes[pindex] > minsize) 
        { 
           size = primes[pindex]; 
           break; 
        }
    }

    h = (hscal_hashtable *)malloc(sizeof(hscal_hashtable));

    if (NULL == h) 
    {
       return NULL;
    }

    h->table = (struct entry **)malloc(sizeof(struct entry*) * size);

    if (NULL == h->table) 
    { 
       free(h); 
       return NULL; 
    }

    memset(h->table, 0, size * sizeof(struct entry *));

    h->tablelength  = size;
    h->primeindex   = pindex;
    h->entrycount   = 0;
    h->hashfn       = hashf;
    h->eqfn         = eqf;
    h->loadlimit    = (unsigned int) ceil(size * max_load_factor);

    return h;
}

/*---------------------------------------------------------------------------*/

unsigned int
hscal_hash(hscal_hashtable *h, void *k)
{
    /* 
    * Aim to protect against poor hash functions by adding logic here
    * - logic taken from java 1.4 hashtable source 
    */
    unsigned int i = h->hashfn(k);

    i += ~(i << 9);
    i ^=  ((i >> 14) | (i << 18)); /* >>> */
    i +=  (i << 4);
    i ^=  ((i >> 10) | (i << 22)); /* >>> */

    return i;
}

/*---------------------------------------------------------------------------*/

int
hscal_hashtable_expand(hscal_hashtable *h)
{
    /* 
    * Double the size of the table to accomodate more entries 
    */
    struct entry **newtable;
    struct entry *e;
    struct entry **pE;
    unsigned int newsize, i, index;

    /* 
    * Check we're not hitting max capacity 
    */
    if (h->primeindex == (prime_table_length - 1)) 
    {
       return 0;
    }

    newsize = primes[++(h->primeindex)];

    newtable = (struct entry **)malloc(sizeof(struct entry*) * newsize);

    if (NULL != newtable)
    {
        memset(newtable, 0, newsize * sizeof(struct entry *));
        /* 
        * This algorithm is not 'stable'. ie. it reverses the list
        * when it transfers entries between the tables 
        */
        for (i = 0; i < h->tablelength; i++) 
        {
            while (NULL != (e = h->table[i])) 
            {
                h->table[i] = e->next;
                index = hscal_indexFor(newsize,e->h);
                e->next = newtable[index];
                newtable[index] = e;
            }
        }

        free(h->table);
        h->table = newtable;
    }
    else 
    {
        /*
        * Plan B: realloc instead
        */
        newtable = (struct entry **)
                   realloc(h->table, newsize * sizeof(struct entry *));

        if (NULL == newtable) 
        { 
           (h->primeindex)--; 
           return 0; 
        }

        h->table = newtable;

        memset(newtable[h->tablelength], 0, newsize - h->tablelength);

        for (i = 0; i < h->tablelength; i++) 
        {
            for (pE = &(newtable[i]), e = *pE; e != NULL; e = *pE) 
            {
                index = hscal_indexFor(newsize,e->h);
                if (index == i)
                {
                    pE = &(e->next);
                }
                else
                {
                    *pE = e->next;
                    e->next = newtable[index];
                    newtable[index] = e;
                }
            }
        }
    }

    h->tablelength = newsize;
    h->loadlimit   = (unsigned int) ceil(newsize * max_load_factor);

    return 1;
}

/*---------------------------------------------------------------------------*/

unsigned int
hscal_hashtable_count(hscal_hashtable *h)
{
    return h->entrycount;
}

/*---------------------------------------------------------------------------*/
int
hscal_hashtable_insert(hscal_hashtable *h, void *k, void *v)
{
    /* 
    * This method allows duplicate keys - but they shouldn't be used 
    */
    unsigned int index;
    struct entry *e;

    if (++(h->entrycount) > h->loadlimit)
    {
        /* Ignore the return value. If expand fails, we should
         * still try cramming just this value into the existing table
         * -- we may not have memory for a larger table, but one more
         * element may be ok. Next time we insert, we'll try expanding again.
         */
        hscal_hashtable_expand(h);
    }

    e = (struct entry *)malloc(sizeof(struct entry));

    if (NULL == e) 
    { 
       --(h->entrycount); 
       return 0; 
    }

    e->h = hscal_hash(h,k);
    index = hscal_indexFor(h->tablelength,e->h);
    e->k = k;
    e->v = v;
    e->next = h->table[index];
    h->table[index] = e;

    return 1;
}

/*---------------------------------------------------------------------------*/
void* 
hscal_hashtable_search(hscal_hashtable *h, void *k)
{
    struct entry *e;
    unsigned int hashvalue, index;

    hashvalue = hscal_hash(h,k);
    index = hscal_indexFor(h->tablelength,hashvalue);
    e = h->table[index];

    while (NULL != e)
    {
        /* 
        * Check hash value to short circuit heavier comparison 
        */
        if ((hashvalue == e->h) && (h->eqfn(k, e->k))) 
        {
           return e->v;
        }

        e = e->next;
    }

    return NULL;
}

/*---------------------------------------------------------------------------*/
void*
hscal_hashtable_remove(hscal_hashtable *h, void *k)
{
    struct entry *e;
    struct entry **pE;
    void *v;
    unsigned int hashvalue, index;

    hashvalue = hscal_hash(h,k);
    index = hscal_indexFor(h->tablelength, hscal_hash(h,k));
    pE = &(h->table[index]);
    e = *pE;

    while (NULL != e)
    {
        /* 
        * Check hash value to short circuit heavier comparison 
        */
        if ((hashvalue == e->h) && (h->eqfn(k, e->k)))
        {
            *pE = e->next;
            h->entrycount--;
            v = e->v;
            free(e);
            return v;
        }
        pE = &(e->next);
        e = e->next;
    }

    return NULL;
}

/*---------------------------------------------------------------------------*/

void
hscal_hashtable_destroy(hscal_hashtable *h, int free_values)
{
    unsigned int i;
    struct entry *e, *f;
    struct entry **table = h->table;

    {
        for (i = 0; i < h->tablelength; i++)
        {
            e = table[i];
            while (NULL != e)
            { f = e; e = e->next; free(f); }
        }
    }

    free(h->table);
    free(h);
}

/*---------------------------------------------------------------------------*/

unsigned int
hscal_hashfromkey(void *ky)
{
   int key = *(int*)ky;

   key -= (key<<6);
   key ^= (key>>17);
   key -= (key<<9);
   key ^= (key<<4);
   key -= (key<<3);
   key ^= (key<<10);
   key ^= (key>>15);

   return key;
}

/*--------------------------------------------------------------------------*/

int
hscal_equalkeys(void *k1, void *k2)
{
    int key1 = *(int*)k1;
    int key2 = *(int*)k2;
    return ((k1 == k2) && (key1 == key2));
}

/*--------------------------------------------------------------------------*/

