
/*******************************************************************
*                                                                  *
*                 Heterogeneous ScaLAPACK Tools                    *
*                 =============================                    *
*                                                                  *
*  Copyright (c) 2006   School of Computer Science and Informatics *
*                              University College Dublin.          *
*                                                                  *
*  All rights reserved. We assume no responsibility for the use    *
*  or reliability of our software.                                 *
*                                                                  *
********************************************************************/

#ifndef __HMPI_SCALAPACK_HASHTABLE_INCLUDE_H_
#define __HMPI_SCALAPACK_HASHTABLE_INCLUDE_H_

/*---------------------------------------------------------------------------*/

struct entry
{
    void *k, *v;
    unsigned int h;
    struct entry *next;
};

/*---------------------------------------------------------------------------*/

struct hashtable {
    unsigned int tablelength;
    struct entry **table;
    unsigned int entrycount;
    unsigned int loadlimit;
    unsigned int primeindex;
    unsigned int (*hashfn) (void *k);
    int (*eqfn) (void *k1, void *k2);
};

/*---------------------------------------------------------------------------*/

unsigned int hscal_hash(
   struct hashtable *h,
   void *k
);

/*---------------------------------------------------------------------------*/

unsigned int hscal_indexFor(
    unsigned int tablelength,
    unsigned int hashvalue
);

/*---------------------------------------------------------------------------*/

typedef struct hashtable hscal_hashtable;

/*
 * create_hashtable
 *  
 * @name                    create_hashtable
 * @param   minsize         minimum initial size of hashtable
 * @param   hashfunction    function for hashing keys
 * @param   key_eq_fn       function for determining key equality
 * @return                  newly created hashtable or NULL on failure
 */
hscal_hashtable *
hscal_create_hashtable(
     unsigned int minsize,
     unsigned int (*hashfunction) (void*),
     int (*key_eq_fn) (void*,void*)
);

/*---------------------------------------------------------------------------*/

/*
 * hscal_hashtable_insert
 *  
 * @name        hashtable_insert
 * @param   h   the hashtable to insert into
 * @param   k   the key - hashtable claims ownership and will free on removal
 * @param   v   the value - does not claim ownership
 * @return      non-zero for successful insertion
 *
 * This function will cause the table to expand if the insertion would take
 * the ratio of entries to table size over the maximum load factor.
 *
 * This function does not check for repeated insertions with a duplicate key.
 * The value returned when using a duplicate key is undefined -- when
 * the hashtable changes size, the order of retrieval of duplicate key
 * entries is reversed.
 * If in doubt, remove before insert.
 */
int 
hscal_hashtable_insert(
     hscal_hashtable *h, 
     void *k, 
     void *v
);

/*---------------------------------------------------------------------------*/

/*
 * hscal_hashtable_search
 *  
 * @name        hashtable_search
 * @param   h   the hashtable to search
 * @param   k   the key to search for  - does not claim ownership
 * @return      the value associated with the key, or NULL if none found
 */
void *
hscal_hashtable_search(
     hscal_hashtable *h, 
     void *k
);

/*---------------------------------------------------------------------------*/

/*
 * hscal_hashtable_remove
 *  
 * @name        hashtable_remove
 * @param   h   the hashtable to remove the item from
 * @param   k   the key to search for  - does not claim ownership
 * @return      the value associated with the key, or NULL if none found
 */
void *
hscal_hashtable_remove(
     hscal_hashtable *h, 
     void *k
);

/*---------------------------------------------------------------------------*/

/*
 * hscal_hashtable_count
 *  
 * @name        hashtable_count
 * @param   h   the hashtable
 * @return      the number of items stored in the hashtable
 */
unsigned int
hscal_hashtable_count(
     hscal_hashtable *h
);

/*---------------------------------------------------------------------------*/

/*
 * hscal_hashtable_destroy
 *
 * @name        hashtable_destroy
 * @param   h               the hashtable
 * @param   free_values     whether to call 'free' on the remaining values
 */
void
hscal_hashtable_destroy(
     hscal_hashtable *h, 
     int free_values
);

/*--------------------------------------------------------------------------*/

unsigned int
hscal_hashfromkey(
     void *ky
);

/*--------------------------------------------------------------------------*/

int
hscal_equalkeys(
     void *k1, 
     void *k2
);

/*--------------------------------------------------------------------------*/


#endif /* __HMPI_SCALAPACK_HASHTABLE_INCLUDE_H_ */

